# coding: utf-8
# pylint: disable=too-many-lines,too-many-public-methods
from datetime import datetime
from functools import lru_cache, partialmethod
from itertools import chain
from operator import attrgetter
from time import sleep

import ujson as json
from first import first

from . import logger
from .cache import AudioFeatures, Playlist, db, db_session, select
from .constants import (
    API,
    DEVICE_ID_RE,
    MANELISTI,
    PLAYLIST_URI_RE,
    AudioFeature,
    TimeRange,
)
from .exceptions import (
    SpotifyAuthException,
    SpotifyDeviceUnavailableException,
    SpotifyException,
    SpotifyForbiddenException,
    SpotifyRateLimitException,
)
from .mixins import AuthMixin, EmailMixin
from .result import SpotifyResult


class SpotifyClient(AuthMixin, EmailMixin):
    def __init__(self, *args, proxies=None, requests_timeout=None, **kwargs):
        """
        Create a Spotify API object.

        :param proxies: Definition of proxies
        :param requests_timeout: Tell Requests to stop waiting for a response after a given number of seconds
        """
        super().__init__(*args, **kwargs)
        self.proxies = proxies
        self.requests_timeout = requests_timeout

    @db_session
    def _increment_api_call_count(self):
        try:
            user = self.user
        except Exception:
            logger.warning("Tried to use an inexistent user: %s", self.user_id)
        else:
            user.api_calls += 1
            user.last_usage_at = datetime.utcnow()

    @staticmethod
    def _check_response(response):
        try:
            response.raise_for_status()
        except:
            exception_params = SpotifyClient.get_exception_params(response)
            if response.status_code == 429 or (
                response.status_code >= 500 and response.status_code < 600
            ):
                raise SpotifyRateLimitException(
                    retry_after=int(response.headers.get("Retry-After", 0)),
                    **exception_params,
                )

            if response.status_code == 403:
                raise SpotifyForbiddenException(**exception_params)

            raise SpotifyException(**exception_params)

    @staticmethod
    def get_exception_params(response):
        return {
            "status_code": response.status_code,
            "url": response.url,
            "headers": response.headers,
            "text": response.text,
        }

    def _internal_call(
        self, method, url, payload, params, headers=None, retries=0, check_202=False
    ):
        logger.debug(url)
        if payload and not isinstance(payload, (bytes, str)):
            payload = json.dumps(payload)
        r = self.session.request(
            method,
            url,
            proxies=self.proxies,
            timeout=self.requests_timeout,
            headers={"Content-Type": "application/json", **(headers or {})},
            data=payload,
            params={k: v for k, v in params.items() if v is not None},
            client_id=self.client_id,
            client_secret=self.client_secret,
        )
        logger.debug("HTTP Status Code: {r.status_code}")
        logger.debug("%s: %s", method, r.url)
        if payload and not isinstance(payload, bytes):
            logger.debug("DATA: %s", payload)

        if check_202 and r.status_code == 202:
            if retries > 0:
                logger.warning(
                    "Device is temporarily unavailable. Retrying in 5 seconds..."
                )
                sleep(5)
                return self._internal_call(
                    method, url, payload, params, headers, retries=retries - 1
                )

            exception_params = self.get_exception_params(r)
            raise SpotifyDeviceUnavailableException(**exception_params)

        try:
            self._check_response(r)
        except SpotifyRateLimitException as exc:
            logger.warning(
                "Reached API rate limit. Retrying in %s seconds...", exc.retry_after
            )
            sleep(exc.retry_after)
            return self._internal_call(method, url, payload, params)

        if self.user_id:
            self._increment_api_call_count()
        if r.text and r.text != "null":
            results = r.json()
            logger.debug("RESP: %s", r.text)
            return SpotifyResult(results, _client=self)

        return None

    def _api_call(self, method, url, args=None, payload=None, headers=None, **kwargs):
        if not self.is_authenticated:
            raise SpotifyAuthException

        retries = kwargs.pop("retries", 0)
        check_202 = kwargs.pop("check_202", False)
        if args:
            kwargs.update(args)

        if "device_id" in kwargs:
            try:
                kwargs["device_id"] = self.get_device_id(kwargs["device_id"])
            except ValueError as e:
                logger.exception(e)

        if not url.startswith("http"):
            url = API.PREFIX.value + url
        return self._internal_call(
            method, url, payload, kwargs, headers, retries, check_202
        )

    def _get(self, url, args=None, payload=None, **kwargs):
        return self._api_call("GET", url, args, payload, **kwargs)

    def _post(self, url, args=None, payload=None, **kwargs):
        return self._api_call("POST", url, args, payload, **kwargs)

    def _delete(self, url, args=None, payload=None, **kwargs):
        return self._api_call("DELETE", url, args, payload, **kwargs)

    def _put(self, url, args=None, payload=None, **kwargs):
        return self._api_call("PUT", url, args, payload, **kwargs)

    def previous(self, result, **kwargs):
        """ returns the previous result given a paged result

            Parameters:
                - result - a previously returned paged result
        """
        if result["previous"]:
            return self._get(result["previous"], **kwargs)

        return None

    def track(self, track_id, **kwargs):
        """ returns a single track given the track's ID, URI or URL

            Parameters:
                - track_id - a spotify URI, URL or ID
        """
        _id = self._get_track_id(track_id)
        # pylint: disable=no-member
        return self._get(API.TRACK.value.format(id=_id), **kwargs)

    def tracks(self, tracks, market="from_token", **kwargs):
        """ returns a list of tracks given a list of track IDs, URIs, or URLs

            Parameters:
                - tracks - a list of spotify URIs, URLs or IDs
                - market - an ISO 3166-1 alpha-2 country code.
        """
        track_list = map(self._get_track_id, tracks)
        return self._get(
            API.TRACKS.value, ids=",".join(track_list), market=market, **kwargs
        )

    def artist(self, artist_id, **kwargs):
        """ returns a single artist given the artist's ID, URI or URL

            Parameters:
                - artist_id - an artist ID, URI or URL
        """
        _id = self._get_artist_id(artist_id)
        # pylint: disable=no-member
        return self._get(API.ARTIST.value.format(id=_id), **kwargs)

    def artists(self, artists, **kwargs):
        """ returns a list of artists given the artist IDs, URIs, or URLs

            Parameters:
                - artists - a list of  artist IDs, URIs or URLs
        """
        artist_list = map(self._get_artist_id, artists)
        return self._get(API.ARTISTS.value, ids=",".join(artist_list), **kwargs)

    def artist_albums(
        self, artist_id, album_type=None, country=None, limit=20, offset=0, **kwargs
    ):
        """ Get Spotify catalog information about an artist's albums

            Parameters:
                - artist_id - the artist ID, URI or URL
                - album_type - 'album', 'single', 'appears_on', 'compilation'
                - country - limit the response to one particular country.
                - limit  - the number of albums to return
                - offset - the index of the first album to return
        """
        _id = self._get_artist_id(artist_id)
        return self._get(
            API.ARTIST_ALBUMS.value.format(id=_id),  # pylint: disable=no-member
            album_type=album_type,
            country=country,
            limit=limit,
            offset=offset,
            **kwargs,
        )

    def artist_top_tracks(self, artist_id, country="US", **kwargs):
        """ Get Spotify catalog information about an artist's top 10 tracks
            by country.

            Parameters:
                - artist_id - the artist ID, URI or URL
                - country - limit the response to one particular country.
        """
        _id = self._get_artist_id(artist_id)
        # pylint: disable=no-member
        return self._get(
            API.ARTIST_TOP_TRACKS.value.format(id=_id), country=country, **kwargs
        )

    @lru_cache(maxsize=128)
    def artist_related_artists(self, artist_id, **kwargs):
        """ Get Spotify catalog information about artists similar to an
            identified artist. Similarity is based on analysis of the
            Spotify community's listening history.

            Parameters:
                - artist_id - the artist ID, URI or URL
        """
        _id = self._get_artist_id(artist_id)
        # pylint: disable=no-member
        return self._get(API.ARTIST_RELATED_ARTISTS.value.format(id=_id), **kwargs)

    def album(self, album_id, **kwargs):
        """ returns a single album given the album's ID, URIs or URL

            Parameters:
                - album_id - the album ID, URI or URL
        """
        _id = self._get_album_id(album_id)
        # pylint: disable=no-member
        return self._get(API.ALBUM.value.format(id=_id), **kwargs)

    def album_tracks(self, album_id, limit=50, offset=0, **kwargs):
        """ Get Spotify catalog information about an album's tracks

            Parameters:
                - album_id - the album ID, URI or URL
                - limit  - the number of items to return
                - offset - the index of the first item to return
        """
        _id = self._get_album_id(album_id)
        # pylint: disable=no-member
        return self._get(
            API.ALBUM_TRACKS.value.format(id=_id), limit=limit, offset=offset, **kwargs
        )

    def albums(self, albums, **kwargs):
        """ returns a list of albums given the album IDs, URIs, or URLs

            Parameters:
                - albums - a list of  album IDs, URIs or URLs
        """
        album_list = map(self._get_album_id, albums)
        return self._get(API.ALBUMS.value, ids=",".join(album_list), **kwargs)

    def search(self, url, q, limit=10, offset=0, market="from_token", **kwargs):
        """ searches for an item

            Parameters:
                - q - the search query
                - limit  - the number of items to return
                - offset - the index of the first item to return
                - type - the type of item to return. One of 'artist', 'album',
                         'track' or 'playlist'
                - market - An ISO 3166-1 alpha-2 country code or the string from_token.
        """
        return self._get(url, q=q, limit=limit, offset=offset, market=market, **kwargs)

    def search_track(self, track, limit=10, offset=0, market="from_token", **kwargs):
        return self.search(
            API.SEARCH_TRACK.value,
            track,
            limit=limit,
            offset=offset,
            market=market,
            **kwargs,
        )

    def search_album(self, album, limit=10, offset=0, market="from_token", **kwargs):
        return self.search(
            API.SEARCH_ALBUM.value,
            album,
            limit=limit,
            offset=offset,
            market=market,
            **kwargs,
        )

    def search_artist(self, artist, limit=10, offset=0, market="from_token", **kwargs):
        return self.search(
            API.SEARCH_ARTIST.value,
            artist,
            limit=limit,
            offset=offset,
            market=market,
            **kwargs,
        )

    def search_playlist(
        self, playlist, limit=10, offset=0, market="from_token", **kwargs
    ):
        return self.search(
            API.SEARCH_PLAYLIST.value,
            playlist,
            limit=limit,
            offset=offset,
            market=market,
            **kwargs,
        )

    def profile(self, user, **kwargs):
        """ Gets basic profile information about a Spotify User

            Parameters:
                - user - the id of the user
        """
        # pylint: disable=no-member
        return self._get(API.USER.value.format(user_id=user), **kwargs)

    def current_user_playlists(self, limit=50, offset=0, **kwargs):
        """ Get current user playlists without required getting his profile
            Parameters:
                - limit  - the number of items to return
                - offset - the index of the first item to return
        """
        return self._get(API.MY_PLAYLISTS.value, limit=limit, offset=offset, **kwargs)

    def user_playlists(self, user, limit=50, offset=0, **kwargs):
        """ Gets playlists of a user

            Parameters:
                - user - the id of the usr
                - limit  - the number of items to return
                - offset - the index of the first item to return
        """
        # pylint: disable=no-member
        return self._get(
            API.PLAYLISTS.value.format(user_id=user),
            limit=limit,
            offset=offset,
            **kwargs,
        )

    def user_playlist(
        self, user, playlist_id=None, fields=None, market="from_token", **kwargs
    ):
        """ Gets playlist of a user
            Parameters:
                - user - the id of the user
                - playlist_id - the id of the playlist
                - fields - which fields to return
        """
        if playlist_id is None:
            return self._get("users/%s/starred" % (user), fields=fields, **kwargs)

        _id = self._get_playlist_id(playlist_id)
        # pylint: disable=no-member
        return self._get(
            API.PLAYLIST.value.format(user_id=user, playlist_id=_id),
            fields=fields,
            market=market,
            **kwargs,
        )

    def user_playlist_tracks(
        self,
        user,
        playlist_id=None,
        fields=None,
        limit=100,
        offset=0,
        market="from_token",
        **kwargs,
    ):
        """ Get full details of the tracks of a playlist owned by a user.

            Parameters:
                - user - the id of the user
                - playlist_id - the id of the playlist
                - fields - which fields to return
                - limit - the maximum number of tracks to return
                - offset - the index of the first track to return
                - market - an ISO 3166-1 alpha-2 country code.
        """
        _id = self._get_playlist_id(playlist_id)
        # pylint: disable=no-member
        return self._get(
            API.PLAYLIST_TRACKS.value.format(user_id=user, playlist_id=_id),
            limit=limit,
            offset=offset,
            fields=fields,
            market=market,
            **kwargs,
        )

    def user_playlist_create(self, user, name, public=True, description="", **kwargs):
        """ Creates a playlist for a user

            Parameters:
                - user - the id of the user
                - name - the name of the playlist
                - public - is the created playlist public
                - description - the description of the playlist
        """
        # pylint: disable=no-member
        data = {"name": name, "public": public, "description": description}
        return self._post(
            API.PLAYLISTS.value.format(user_id=user), payload=data, **kwargs
        )

    def user_playlist_upload_cover_image(self, user, playlist_id, image, **kwargs):
        """ Creates a playlist for a user

            Parameters:
                - user - the id of the user
                - playlist_id - the id of the playlist
                - image - base64 encoded image
        """
        # pylint: disable=no-member
        return self._put(
            API.PLAYLIST_IMAGES.value.format(user_id=user, playlist_id=playlist_id),
            payload=image,
            headers={"Content-Type": "image/jpeg"},
            **kwargs,
        )

    def user_playlist_change_details(
        self,
        user,
        playlist_id,
        name=None,
        public=None,
        collaborative=None,
        description=None,
        **kwargs,
    ):
        """ Changes a playlist's name and/or public/private state

            Parameters:
                - user - the id of the user
                - playlist_id - the id of the playlist
                - name - optional name of the playlist
                - public - optional is the playlist public
                - collaborative - optional is the playlist collaborative
                - description - the description of the playlist
        """
        data = {}
        if isinstance(name, str):
            data["name"] = name
        if isinstance(public, bool):
            data["public"] = public
        if isinstance(collaborative, bool):
            data["collaborative"] = collaborative
        if isinstance(collaborative, str):
            data["description"] = description
        # pylint: disable=no-member
        return self._put(
            API.PLAYLIST.value.format(user_id=user, playlist_id=playlist_id),
            payload=data,
            **kwargs,
        )

    def user_playlist_unfollow(self, user, playlist_id, **kwargs):
        """ Unfollows (deletes) a playlist for a user

            Parameters:
                - user - the id of the user
                - name - the name of the playlist
        """
        return self._delete(
            "users/%s/playlists/%s/followers" % (user, playlist_id), **kwargs
        )

    def user_playlist_add_tracks(
        self, user, playlist_id, tracks, position=None, **kwargs
    ):
        """ Adds tracks to a playlist

            Parameters:
                - user - the id of the user
                - playlist_id - the id of the playlist
                - tracks - a list of track URIs, URLs or IDs
                - position - the position to add the tracks
        """
        _id = self._get_playlist_id(playlist_id)
        # pylint: disable=no-member
        url = API.PLAYLIST_TRACKS.value.format(user_id=user, playlist_id=_id)
        track_uris = list(map(self._get_track_uri, tracks))
        if len(track_uris) <= 100:
            return self._post(
                url, payload={"uris": track_uris}, position=position, **kwargs
            )

        batches = [
            {"uris": track_uris[i : i + 100]} for i in range(0, len(track_uris), 100)
        ]
        results = [
            self._post(
                url,
                payload=t,
                position=(None if position is None else i * 100 + position),
                **kwargs,
            )
            for i, t in enumerate(batches)
        ]
        return results

    def user_playlist_replace_tracks(self, user, playlist_id, tracks, **kwargs):
        """ Replace all tracks in a playlist

            Parameters:
                - user - the id of the user
                - playlist_id - the id of the playlist
                - tracks - the list of track ids to add to the playlist
        """
        _id = self._get_playlist_id(playlist_id)
        # pylint: disable=no-member
        url = API.PLAYLIST_TRACKS.value.format(user_id=user, playlist_id=_id)
        first_100_tracks, rest_tracks = tracks[:100], tracks[100:]
        track_uris = list(map(self._get_track_uri, first_100_tracks))
        replaced = self._put(url, payload={"uris": track_uris}, **kwargs)
        if not rest_tracks:
            return replaced

        added = self.user_playlist_add_tracks(user, playlist_id, rest_tracks)
        if isinstance(added, list):
            return [replaced, *added]

        return [replaced, added]

    def user_playlist_reorder_tracks(
        self,
        user,
        playlist_id,
        range_start,
        insert_before,
        range_length=1,
        snapshot_id=None,
        **kwargs,
    ):
        """ Reorder tracks in a playlist

            Parameters:
                - user - the id of the user
                - playlist_id - the id of the playlist
                - range_start - the position of the first track to be reordered
                - range_length - optional the number of tracks to be reordered (default: 1)
                - insert_before - the position where the tracks should be inserted
                - snapshot_id - optional playlist's snapshot ID
        """
        _id = self._get_playlist_id(playlist_id)
        payload = {
            "range_start": range_start,
            "range_length": range_length,
            "insert_before": insert_before,
        }
        if snapshot_id:
            payload["snapshot_id"] = snapshot_id
        # pylint: disable=no-member
        return self._put(
            API.PLAYLIST_TRACKS.value.format(user_id=user, playlist_id=_id),
            payload=payload,
            **kwargs,
        )

    def user_playlist_remove_all_occurrences_of_tracks(
        self, user, playlist_id, tracks, snapshot_id=None, **kwargs
    ):
        """ Removes all occurrences of the given tracks from the given playlist

            Parameters:
                - user - the id of the user
                - playlist_id - the id of the playlist
                - tracks - the list of track ids to add to the playlist
                - snapshot_id - optional id of the playlist snapshot

        """
        _id = self._get_playlist_id(playlist_id)
        track_uris = map(self._get_track_uri, tracks)
        payload = {"tracks": [{"uri": track} for track in track_uris]}
        if snapshot_id:
            payload["snapshot_id"] = snapshot_id
        # pylint: disable=no-member
        return self._delete(
            API.PLAYLIST_TRACKS.value.format(user_id=user, playlist_id=_id),
            payload=payload,
            **kwargs,
        )

    def user_playlist_remove_specific_occurrences_of_tracks(
        self, user, playlist_id, tracks, snapshot_id=None, **kwargs
    ):
        """ Removes all occurrences of the given tracks from the given playlist

            Parameters:
                - user - the id of the user
                - playlist_id - the id of the playlist
                - tracks - an array of objects containing Spotify URIs of the tracks
                           to remove with their current positions in the playlist.
                    For example:
                        [
                          { "uri":"4iV5W9uYEdYUVa79Axb7Rh", "positions":[2] },
                          { "uri":"1301WleyT98MSxVHPZCA6M", "positions":[7] }
                        ]
                - snapshot_id - optional id of the playlist snapshot
        """
        _id = self._get_playlist_id(playlist_id)
        ftracks = []
        for tr in tracks:
            ftracks.append(
                {"uri": self._get_uri("track", tr["uri"]), "positions": tr["positions"]}
            )
        payload = {"tracks": ftracks}
        if snapshot_id:
            payload["snapshot_id"] = snapshot_id
        # pylint: disable=no-member
        return self._delete(
            API.PLAYLIST_TRACKS.value.format(user_id=user, playlist_id=_id),
            payload=payload,
            **kwargs,
        )

    def user_playlist_follow_playlist(self, playlist_owner_id, playlist_id, **kwargs):
        """
        Add the current authenticated user as a follower of a playlist.

        Parameters:
            - playlist_owner_id - the user id of the playlist owner
            - playlist_id - the id of the playlist

        """
        return self._put(
            API.PLAYLIST_FOLLOWERS.value.format(  # pylint: disable=no-member
                owner_id=playlist_owner_id, playlist_id=playlist_id
            ),
            **kwargs,
        )

    def user_playlist_is_following(
        self, playlist_owner_id, playlist_id, user_ids, **kwargs
    ):
        """
        Check to see if the given users are following the given playlist

        Parameters:
            - playlist_owner_id - the user id of the playlist owner
            - playlist_id - the id of the playlist
            - user_ids - the ids of the users that you want to check to see if they follow the playlist. Maximum: 5 ids.

        """
        return self._get(
            API.PLAYLIST_FOLLOWERS_CONTAINS.value.format(  # pylint: disable=no-member
                user_id=playlist_owner_id, playlist_id=playlist_id
            ),
            ids=",".join(user_ids),
            **kwargs,
        )

    def me(self, **kwargs):
        """ Get detailed profile information about the current user.
            An alias for the 'current_user' method.
        """
        return self._get(API.ME.value, **kwargs)

    def current_user(self, **kwargs):
        """ Get detailed profile information about the current user.
            An alias for the 'me' method.
        """
        return self.me(**kwargs)

    def current_user_saved_albums(self, limit=20, offset=0, **kwargs):
        """ Gets a list of the albums saved in the current authorized user's
            "Your Music" library

            Parameters:
                - limit - the number of albums to return
                - offset - the index of the first album to return

        """
        return self._get(API.MY_ALBUMS.value, limit=limit, offset=offset, **kwargs)

    def current_user_saved_tracks(self, limit=20, offset=0, **kwargs):
        """ Gets a list of the tracks saved in the current authorized user's
            "Your Music" library

            Parameters:
                - limit - the number of tracks to return
                - offset - the index of the first track to return

        """
        return self._get(API.MY_TRACKS.value, limit=limit, offset=offset, **kwargs)

    def current_user_followed_artists(self, limit=20, after=None, **kwargs):
        """ Gets a list of the artists followed by the current authorized user

            Parameters:
                - limit - the number of tracks to return
                - after - ghe last artist ID retrieved from the previous request

        """
        return self._get(
            API.MY_FOLLOWING.value, type="artist", limit=limit, after=after, **kwargs
        )

    def user_follow_artists(self, ids=None, **kwargs):
        """ Follow one or more artists
            Parameters:
                - ids - a list of artist IDs
        """
        return self._put(
            API.MY_FOLLOWING.value, type="artist", ids=",".join(ids or []), **kwargs
        )

    def user_follow_users(self, ids=None, **kwargs):
        """ Follow one or more users
            Parameters:
                - ids - a list of user IDs
        """
        return self._put(
            API.MY_FOLLOWING.value, type="user", ids=",".join(ids or []), **kwargs
        )

    def current_user_saved_tracks_delete(self, tracks=None, **kwargs):
        """ Remove one or more tracks from the current user's
            "Your Music" library.

            Parameters:
                - tracks - a list of track URIs, URLs or IDs
        """
        track_list = []
        if tracks is not None:
            track_list = map(self._get_track_id, tracks)
        return self._delete(API.MY_TRACKS.value, ids=",".join(track_list), **kwargs)

    def current_user_saved_tracks_contains(self, tracks=None, **kwargs):
        """ Check if one or more tracks is already saved in
            the current Spotify user’s “Your Music” library.

            Parameters:
                - tracks - a list of track URIs, URLs or IDs
        """
        track_list = []
        if tracks is not None:
            track_list = map(self._get_track_id, tracks)
        return self._get(API.MY_TRACKS_CONTAINS.value, ",".join(track_list), **kwargs)

    def current_user_saved_tracks_add(self, tracks=None, **kwargs):
        """ Add one or more tracks to the current user's
            "Your Music" library.

            Parameters:
                - tracks - a list of track URIs, URLs or IDs
        """
        track_list = []
        if tracks is not None:
            track_list = map(self._get_track_id, tracks)
        return self._put(API.MY_TRACKS.value, ids=",".join(track_list), **kwargs)

    def current_user_top_artists(
        self, limit=20, offset=0, time_range=TimeRange.MEDIUM_TERM, **kwargs
    ):
        """ Get the current user's top artists

            Parameters:
                - limit - the number of entities to return
                - offset - the index of the first entity to return
                - time_range - Over what time frame are the affinities computed
                  Valid-values: short_term, medium_term, long_term
        """
        return self._get(
            API.MY_TOP.value.format(type="artists"),  # pylint: disable=no-member
            time_range=TimeRange(time_range).value,
            limit=limit,
            offset=offset,
            **kwargs,
        )

    def current_user_top_tracks(
        self, limit=20, offset=0, time_range=TimeRange.MEDIUM_TERM, **kwargs
    ):
        """ Get the current user's top tracks

            Parameters:
                - limit - the number of entities to return
                - offset - the index of the first entity to return
                - time_range - Over what time frame are the affinities computed
                  Valid-values: short_term, medium_term, long_term
        """
        return self._get(
            API.MY_TOP.value.format(type="tracks"),  # pylint: disable=no-member
            time_range=TimeRange(time_range).value,
            limit=limit,
            offset=offset,
            **kwargs,
        )

    def current_user_saved_albums_add(self, albums=None, **kwargs):
        """ Add one or more albums to the current user's
            "Your Music" library.
            Parameters:
                - albums - a list of album URIs, URLs or IDs
        """
        album_list = map(self._get_album_id, albums or [])
        return self._put(API.MY_ALBUMS.value, ids=",".join(album_list), **kwargs)

    def featured_playlists(
        self, locale=None, country=None, timestamp=None, limit=20, offset=0, **kwargs
    ):
        """ Get a list of Spotify featured playlists

            Parameters:
                - locale - The desired language, consisting of a lowercase ISO
                  639 language code and an uppercase ISO 3166-1 alpha-2 country
                  code, joined by an underscore.

                - country - An ISO 3166-1 alpha-2 country code.

                - timestamp - A timestamp in ISO 8601 format:
                  yyyy-MM-ddTHH:mm:ss. Use this parameter to specify the user's
                  local time to get results tailored for that specific date and
                  time in the day

                - limit - The maximum number of items to return. Default: 20.
                  Minimum: 1. Maximum: 50

                - offset - The index of the first item to return. Default: 0
                  (the first object). Use with limit to get the next set of
                  items.
        """
        return self._get(
            API.FEATURED_PLAYLISTS.value,
            locale=locale,
            country=country,
            timestamp=timestamp,
            limit=limit,
            offset=offset,
            **kwargs,
        )

    def new_releases(self, country=None, limit=20, offset=0, **kwargs):
        """ Get a list of new album releases featured in Spotify

            Parameters:
                - country - An ISO 3166-1 alpha-2 country code.

                - limit - The maximum number of items to return. Default: 20.
                  Minimum: 1. Maximum: 50

                - offset - The index of the first item to return. Default: 0
                  (the first object). Use with limit to get the next set of
                  items.
        """
        return self._get(
            API.NEW_RELEASES.value,
            country=country,
            limit=limit,
            offset=offset,
            **kwargs,
        )

    def categories(self, country=None, locale=None, limit=20, offset=0, **kwargs):
        """ Get a list of new album releases featured in Spotify

            Parameters:
                - country - An ISO 3166-1 alpha-2 country code.
                - locale - The desired language, consisting of an ISO 639
                  language code and an ISO 3166-1 alpha-2 country code, joined
                  by an underscore.

                - limit - The maximum number of items to return. Default: 20.
                  Minimum: 1. Maximum: 50

                - offset - The index of the first item to return. Default: 0
                  (the first object). Use with limit to get the next set of
                  items.
        """
        return self._get(
            API.CATEGORIES.value,
            country=country,
            locale=locale,
            limit=limit,
            offset=offset,
            **kwargs,
        )

    def category_playlists(
        self, category_id=None, country=None, limit=20, offset=0, **kwargs
    ):
        """ Get a list of new album releases featured in Spotify

            Parameters:
                - category_id - The Spotify category ID for the category.

                - country - An ISO 3166-1 alpha-2 country code.

                - limit - The maximum number of items to return. Default: 20.
                  Minimum: 1. Maximum: 50

                - offset - The index of the first item to return. Default: 0
                  (the first object). Use with limit to get the next set of
                  items.
        """
        # pylint: disable=no-member
        return self._get(
            API.CATEGORY_PLAYLISTS.value.format(id=category_id),
            country=country,
            limit=limit,
            offset=offset,
            **kwargs,
        )

    # pylint: disable=too-many-locals
    def recommendations(
        self,
        seed_artists=None,
        seed_genres=None,
        seed_tracks=None,
        limit=20,
        country="from_token",
        filter_manele=True,
        **kwargs,
    ):
        """ Get a list of recommended tracks for one to five seeds.

            Parameters:
                - seed_artists - a list of artist IDs, URIs or URLs

                - seed_tracks - a list of artist IDs, URIs or URLs

                - seed_genres - a list of genre names. Available genres for
                  recommendations can be found by calling recommendation_genre_seeds

                - country - An ISO 3166-1 alpha-2 country code. If provided, all
                  results will be playable in this country.

                - limit - The maximum number of items to return. Default: 20.
                  Minimum: 1. Maximum: 100

                - min/max/target_<attribute> - For the tuneable track attributes listed
                  in the documentation, these values provide filters and targeting on
                  results.
        """
        params = dict(limit=limit)
        if seed_artists:
            params["seed_artists"] = ",".join(map(self._get_artist_id, seed_artists))
        if seed_genres:
            params["seed_genres"] = ",".join(seed_genres)
        if seed_tracks:
            params["seed_tracks"] = ",".join(map(self._get_track_id, seed_tracks))
        if country:
            params["market"] = country
        for attribute in list(AudioFeature):
            for prefix in ["min_", "max_", "target_"]:
                param = prefix + attribute.value
                if param in kwargs:
                    params[param] = kwargs.pop(param)

        if not filter_manele:
            return self._get(API.RECOMMENDATIONS.value, **params, **kwargs)

        for _ in range(5):
            result = self._get(API.RECOMMENDATIONS.value, **params, **kwargs)
            tracks = [
                t
                for t in result.tracks
                if not any(
                    "manele" in (a.genres or []) or a.id in MANELISTI for a in t.artists
                )
            ]
            if tracks:
                result.tracks = tracks
                return result
        return self._get(API.RECOMMENDATIONS.value, **params, **kwargs)

    def recommendation_genre_seeds(self, **kwargs):
        """ Get a list of genres available for the recommendations function.
        """
        return self._get(API.RECOMMENDATIONS_GENRES.value, **kwargs)

    def audio_analysis(self, track=None, **kwargs):
        """ Get audio analysis for a track based upon its Spotify ID
            Parameters:
                - track - a track URI, URL or ID
        """
        _id = self._get_track_id(track)
        # pylint: disable=no-member
        return self._get(API.AUDIO_ANALYSIS.value.format(id=_id), **kwargs)

    def audio_features(self, track=None, tracks=None, with_cache=True, **kwargs):
        """ Get audio features for one or multiple tracks based upon their Spotify IDs
            Parameters:
                - track - a track URI, URL or ID
                - tracks - a list of track URIs, URLs or IDs, maximum: 100 ids
        """
        if track:
            _id = self._get_track_id(track)
            # pylint: disable=no-member
            return self._get(API.AUDIO_FEATURES_SINGLE.value.format(id=_id), **kwargs)

        tracks = list(map(self._get_track_id, tracks or []))
        cached_tracks = []
        if with_cache:
            with db_session:
                cached_tracks = select(a for a in AudioFeatures if a.id in tracks)[:]
                tracks = list(set(tracks) - {a.id for a in cached_tracks})
        batches = [tracks[i : i + 100] for i in range(0, len(tracks), 100)]
        audio_features = [
            self._get(API.AUDIO_FEATURES_MULTIPLE.value, ids=",".join(t), **kwargs)
            for t in batches
        ]
        with db_session:
            audio_features = [
                AudioFeatures.from_dict(t) for t in chain.from_iterable(audio_features)
            ] + cached_tracks
        return audio_features

    def devices(self, **kwargs):
        """ Get a list of user's available devices.
        """
        return self._get(API.DEVICES.value, check_202=True, **kwargs)

    def get_device_id(self, device=None):
        if isinstance(device, (str, bytes)) and DEVICE_ID_RE.match(device):
            return device

        return self.get_device(device).id

    def get_device(self, device=None, only_active=True):
        """Get Spotify device based on name

        :param str, optional device: device name or ID
        :param str, optional field: device attribute to return

        str or dict: Spotify device
        """
        devices = self.devices().devices
        device_names = ", ".join([d.name for d in devices])
        device_name_or_id = device
        if not device_name_or_id:
            if only_active:
                device = first(devices, key=attrgetter("is_active"))
                if not device:
                    raise ValueError(
                        f"""
            There's no active device.
            Possible devices: {device_names}"""
                    )
            else:
                device = first(devices, key=attrgetter("is_active")) or first(devices)
        else:
            device = first(devices, key=lambda d: device_name_or_id in (d.name, d.id))
            if not device:
                raise ValueError(
                    f"""
        Device {device_name_or_id} doesn't exist.
        Possible devices: {device_names}"""
                )

        return device

    def current_playback(self, market="from_token", **kwargs):
        """ Get information about user's current playback.

            Parameters:
                - market - an ISO 3166-1 alpha-2 country code.
        """
        return self._get(API.PLAYER.value, market=market, **kwargs)

    def current_user_recently_played(self, limit=50, **kwargs):
        """ Get the current user's recently played tracks

            Parameters:
                - limit - the number of entities to return
        """
        return self._get(API.RECENTLY_PLAYED.value, limit=limit, **kwargs)

    def currently_playing(self, market="from_token", **kwargs):
        """ Get user's currently playing track.

            Parameters:
                - market - an ISO 3166-1 alpha-2 country code.
        """
        return self._get(
            API.CURRENTLY_PLAYING.value, market=market, check_202=True, **kwargs
        )

    def transfer_playback(self, device, force_play=True, **kwargs):
        """ Transfer playback to another device.
            Note that the API accepts a list of device ids, but only
            actually supports one.

            Parameters:
                - device - transfer playback to this device
                - force_play - true: after transfer, play. false:
                               keep current state.
        """
        device_id = self.get_device_id(device)
        data = {"device_ids": [device_id], "play": force_play}
        return self._put(API.PLAYER.value, payload=data, check_202=True, **kwargs)

    def start_playback(
        self,
        device=None,
        artist=None,
        album=None,
        playlist=None,
        tracks=None,
        offset=None,
        **kwargs,
    ):
        """ Start or resume user's playback.

            Parameters:
                - device - device target for playback
                - playlist - spotify playlist to play
                - artist - spotify artist to play
                - album - spotify album to play
                - tracks - spotify tracks to play
                - offset - offset into context by index or track
        """
        data = {}
        if playlist:
            data["context_uri"] = self._get_playlist_uri(playlist)
        elif album:
            data["context_uri"] = self._get_album_uri(album)
        elif artist:
            data["context_uri"] = self._get_artist_uri(artist)
        elif tracks:
            data["uris"] = list(map(self._get_track_uri, tracks))
        if isinstance(offset, int):
            data["offset"] = dict(position=offset)
        elif isinstance(offset, str):
            data["offset"] = dict(uri=offset)
        return self._put(
            API.PLAY.value, device_id=device, payload=data, check_202=True, **kwargs
        )

    def pause_playback(self, device=None, **kwargs):
        """ Pause user's playback.

            Parameters:
                - device - device target for playback
        """
        return self._put(API.PAUSE.value, device_id=device, check_202=True, **kwargs)

    def next_track(self, device=None, **kwargs):
        """ Skip user's playback to next track.

            Parameters:
                - device - device target for playback
        """
        return self._post(API.NEXT.value, device_id=device, check_202=True, **kwargs)

    def previous_track(self, device=None, **kwargs):
        """ Skip user's playback to previous track.

            Parameters:
                - device - device target for playback
        """
        return self._post(
            API.PREVIOUS.value, device_id=device, check_202=True, **kwargs
        )

    def seek_track(self, position_ms, device=None, **kwargs):
        """ Seek to position in current track.

            Parameters:
                - position_ms - position in milliseconds to seek to
                - device - device target for playback
        """
        if not isinstance(position_ms, int):
            logger.warning("position_ms must be an integer")
            return None

        return self._put(
            API.SEEK.value,
            position_ms=position_ms,
            device_id=device,
            check_202=True,
            **kwargs,
        )

    def repeat(self, state, device=None, **kwargs):
        """ Set repeat mode for playback.

            Parameters:
                - state - `track`, `context`, or `off`
                - device - device target for playback
        """
        if state not in ["track", "context", "off"]:
            logger.warning("Invalid state")
            return

        self._put(
            API.REPEAT.value, state=state, device_id=device, check_202=True, **kwargs
        )

    def volume(self, volume_percent: int = None, device: str = None, **kwargs):
        """ Get or set playback volume.

            Parameters:
                - volume_percent - volume between 0 and 100
                - device - device target for playback
        """
        device = self.get_device(device)
        if volume_percent is None:
            return device.volume_percent

        assert 0 <= volume_percent <= 100
        return self._put(
            API.VOLUME.value,
            volume_percent=volume_percent,
            device_id=device.id,
            check_202=True,
            **kwargs,
        )

    def shuffle(self, state, device=None, **kwargs):
        """ Toggle playback shuffling.

            Parameters:
                - state - true or false
                - device - device target for playback
        """
        if not isinstance(state, bool):
            logger.warning("State must be a boolean")
            return

        state = str(state).lower()
        self._put(
            API.SHUFFLE.value, state=state, device_id=device, check_202=True, **kwargs
        )

    @staticmethod
    def _get_id(_type, result):
        if isinstance(result, str):
            fields = result.split(":")
            if len(fields) >= 3:
                if _type != fields[-2]:
                    logger.warning(
                        "Expected id of type %s but found type %s %s",
                        _type,
                        fields[-2],
                        result,
                    )
                return fields[-1]

            fields = result.split("/")
            if len(fields) >= 3:
                itype = fields[-2]
                if _type != itype:
                    logger.warning(
                        "Expected id of type %s but found type %s %s",
                        _type,
                        itype,
                        result,
                    )
                return fields[-1]

        elif isinstance(result, (SpotifyResult, db.Entity)):
            return result.id

        elif isinstance(result, dict):
            return result["id"]

        return result

    _get_track_id = partialmethod(_get_id, "track")
    _get_artist_id = partialmethod(_get_id, "artist")
    _get_album_id = partialmethod(_get_id, "album")
    _get_playlist_id = partialmethod(_get_id, "playlist")

    def _get_uri(self, _type, result):
        return "spotify:" + _type + ":" + self._get_id(_type, result)

    def _get_playlist_uri(self, playlist, user=None):
        if isinstance(playlist, (str, bytes)) and PLAYLIST_URI_RE.match(playlist):
            return playlist

        if isinstance(playlist, Playlist):
            return playlist.uri

        if user is not None:
            return f'spotify:user:{self._get_id("user", user)}:playlist:{self._get_id("playlist", playlist)}'

        try:
            if "uri" in playlist:
                return playlist["uri"]

        except:
            pass

        return None

    _get_track_uri = partialmethod(_get_uri, "track")
    _get_artist_uri = partialmethod(_get_uri, "artist")
    _get_album_uri = partialmethod(_get_uri, "album")



import functools
from collections import OrderedDict

from .db import *


def async_lru(maxsize=100):
    cache = OrderedDict()

    def decorator(fn):
        @functools.wraps(fn)
        async def memoizer(*args, **kwargs):
            key = str((args, kwargs))
            try:
                cache[key] = cache.pop(key)
            except KeyError:
                if len(cache) >= maxsize:
                    cache.popitem(last=False)
                cache[key] = await fn(*args, **kwargs)
            return cache[key]

        return memoizer

    return decorator



import random
import time
from datetime import date, timedelta
from itertools import chain

from pony.orm import get
from pony.orm.core import CacheIndexError

from .. import logger
from ..cache import Artist, Genre, Playlist, db_session, select
from ..constants import TimeRange
from ..util import normalize_features


class RecommenderMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @db_session
    def fetch_playlists(self):
        fetched_ids = set(select(p.id for p in Playlist))
        results = chain(
            self.user_playlists("particledetector").all(),
            self.user_playlists("thesoundsofspotify").all(),
        )
        for playlist in results:
            logger.info("Fetching %s", playlist.name)
            if playlist.id not in fetched_ids:
                Playlist.from_dict(playlist)
            fetched_ids.add(playlist.id)

    @db_session
    def fetch_user_top(self, time_range):
        self.user.top_artists.clear()
        for artist in self.current_user_top_artists(
            limit=50, time_range=time_range
        ).iterall():
            if self.is_disliked_artist(artist):
                continue

            if Artist.exists(id=artist.id):
                self.user.top_artists.add(Artist[artist.id])
            else:
                try:
                    artist = Artist.from_dict(artist)
                except CacheIndexError:
                    artist = Artist[artist.id]
                self.user.top_artists.add(artist)
        self.user.top_genres = (
            self.user.top_artists.genres.distinct().keys() - self.user.disliked_genres
        )
        if self.user.top_expires_at is None:
            self.user.top_expires_at = {}
        self.user.top_expires_at[TimeRange(time_range).value] = time.mktime(
            (date.today() + timedelta(days=1)).timetuple()
        )

    @db_session
    def genre_playlist(
        self, genre, popularity=Playlist.Popularity.SOUND
    ):  # pylint: disable=no-self-use
        if not Playlist.exists():
            raise Exception('You have to call "fetch_playlists" first.')

        return Playlist.get(
            genre=genre, popularity=Playlist.Popularity(popularity).value
        )

    @db_session
    def top_artists(self, time_range=TimeRange.SHORT_TERM):
        if self.user.top_expired(time_range):
            self.fetch_user_top(time_range)
        return self.user.top_artists

    @db_session
    def top_genres(self, time_range=TimeRange.SHORT_TERM):
        if self.user.top_expired(time_range):
            self.fetch_user_top(time_range)
        return self.user.top_genres

    def order_by(self, features, tracks):
        audio_features = self.audio_features(tracks=tracks)
        track_ids = [a.id for a in audio_features]
        audio_features = [a.to_dict(list(features.keys())) for a in audio_features]
        audio_features = normalize_features(audio_features, track_ids)
        for feature, direction in features.items():
            audio_features[feature] *= direction
        audio_features["total"] = audio_features.sum(axis=1)
        return audio_features.sort_values("total").index.tolist()

    def fill_with_related_artists(self, artists, limit=5):
        tries = 5
        if len(artists) >= limit:
            return artists

        artist_set = set(artists)
        artist_list = list(artists)
        while len(artist_set) < limit and tries:
            tries -= 1
            related_artists = {
                a.id
                for a in self.artist_related_artists(random.choice(artist_list))
                if self.is_not_disliked_artist(a)
            }
            related_artists_limit = min(
                random.randint(1, limit - len(artist_set)), len(related_artists)
            )
            artist_set |= set(random.sample(related_artists, related_artists_limit))
        return artist_set

    @db_session
    def recommend_by_top_artists(
        self,
        artist_limit=2,
        track_limit=100,
        use_related=True,
        features_order=None,
        time_range=TimeRange.SHORT_TERM,
        **kwargs,
    ):
        """Get a list of recommended songs.

        Returns:
            list: List of tracks
        """
        artists = (
            self.top_artists(time_range=time_range)
            .select()
            .without_distinct()
            .random(artist_limit)
        )
        if use_related:
            artists = self.fill_with_related_artists([a.id for a in artists])
        tracks = self.recommendations(seed_artists=artists, limit=track_limit, **kwargs)
        tracks = list(filter(self.is_not_disliked_track, tracks))
        if features_order:
            tracks = self.order_by(features_order, tracks)
        return tracks

    @db_session
    def is_disliked_artist(self, artist):
        return artist.id in set(
            self.user.disliked_artists.id.distinct().keys()
        ) or bool(set(artist.genres or []) & set(self.user.disliked_genres))

    @db_session
    def is_not_disliked_artist(self, artist):
        return not self.is_disliked_artist(artist)

    @db_session
    def is_not_disliked_track(self, track):
        return all(self.is_not_disliked_artist(a) for a in track.artists)

    @db_session
    def disliked_artists(self):
        return list(self.user.disliked_artists)

    @db_session
    def disliked_genres(self):
        return list(self.user.disliked_genres)

    @db_session
    def dislike_artist(self, artist):
        if isinstance(artist, str):
            # pylint: disable=consider-using-in
            artist = get(a for a in Artist if a.id == artist or a.name == artist)
        elif isinstance(artist, dict):
            artist = Artist.get(artist.id) or Artist.from_dict(artist)
        self.user.dislike(artist=artist)

    @db_session
    def dislike_genre(self, genre):
        if isinstance(genre, str):
            genre = genre.lower()
            genre = Genre.get(name=genre) or Genre(name=genre)
        self.user.dislike(genre=genre)



import asyncio
import random
import sys
from collections import OrderedDict
from functools import partial

from cached_property import cached_property
from first import first

from ... import config
from ...cache import Playlist, db_session
from ...constants import ItemType, TimeRange, VolumeBackend
from ...volume import (
    AlsaVolumeControl,
    ApplescriptVolumeControl,
    LinuxVolumeControlAsync,
    SpotifyVolumeControlAsync,
)


class PlayerMixin:
    def __init__(
        self,
        *args,
        device=None,
        alsa_device=None,
        alsa_mixer=None,
        speaker=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.device = device or config.player.device
        self.alsa_device = alsa_device or config.player.alsa.device
        self.alsa_mixer = alsa_mixer or config.player.alsa.mixer
        self.speaker = speaker or config.player.speaker

    @cached_property
    def _optimal_backend(self):
        return first(self._backends.values())

    @cached_property
    def _backends(self):
        return OrderedDict(
            {
                VolumeBackend.APPLESCRIPT: self._applescript_volume_control,
                VolumeBackend.LINUX: self._linux_volume_control,
                VolumeBackend.ALSA: self._alsa_volume_control,
                VolumeBackend.SPOTIFY: self._spotify_volume_control,
            }
        )

    @cached_property
    def _applescript_volume_control(self):
        try:
            # pylint: disable=no-member
            if sys.platform != "darwin":
                return None

            return ApplescriptVolumeControl(self.speaker)

        except:
            return None

    @cached_property
    def _linux_volume_control(self):
        try:
            # pylint: disable=no-member
            if sys.platform != "linux":
                return None

            return LinuxVolumeControlAsync(
                self,
                self.alsa_mixer,
                spotify_device=self.device,
                alsa_device=self.alsa_device,
            )

        except:
            return None

    @cached_property
    def _alsa_volume_control(self):
        try:
            # pylint: disable=no-member
            if sys.platform != "linux":
                return None

            return AlsaVolumeControl(self.alsa_mixer, device=self.alsa_device)

        except:
            return None

    @cached_property
    def _spotify_volume_control(self):
        return SpotifyVolumeControlAsync(self, device=self.device)

    def backend(self, backend=None, device=None):
        if not backend:
            return self._optimal_backend

        volume_backend = self._backends[VolumeBackend(backend)]
        if not volume_backend:
            raise ValueError(
                f"Backend {volume_backend} is not available on this system"
            )

        if (
            isinstance(volume_backend, SpotifyVolumeControlAsync)
            and device
            and device != self.device
        ):
            volume_backend = SpotifyVolumeControlAsync(self, device=device)
        return volume_backend

    async def change_volume(self, by=0, to=None, backend=None, device=None):
        volume_backend = self.backend(backend, device=device)
        if isinstance(volume_backend, SpotifyVolumeControlAsync):
            if to is not None:
                volume = to
            else:
                volume = await volume_backend.volume()
            await volume_backend.set_volume(volume + by)
        else:
            if to is not None:
                volume = to
            else:
                volume = volume_backend.volume
            volume_backend.volume = volume + by
        return volume

    async def volume_up(self, backend=None):
        return await self.change_volume(by=+1, backend=backend)

    async def volume_down(self, backend=None):
        return await self.change_volume(by=-1, backend=backend)

    async def fade_up(self, **kwargs):
        await self.fade(**{**config.volume.fade.up, **kwargs})

    async def fade_down(self, **kwargs):
        await self.fade(**{**config.volume.fade.down, **kwargs})

    #  pylint: disable=too-many-arguments

    async def fade(
        self,
        limit=50,
        start=0,
        step=1,
        seconds=300,
        force=False,
        backend=None,
        spotify_volume=100,
        device=None,
        blocking=False,
    ):
        volume_backend = self.backend(backend, device=device)
        if not isinstance(volume_backend, SpotifyVolumeControlAsync):
            await self.change_volume(
                to=spotify_volume, backend=VolumeBackend.SPOTIFY, device=device
            )
        await self.change_volume(to=start, backend=backend, device=device)
        kwargs = dict(
            limit=int(limit),
            start=int(start),
            step=int(step),
            seconds=int(seconds),
            force=bool(force),
        )
        loop = asyncio.get_event_loop()
        if isinstance(
            volume_backend, (SpotifyVolumeControlAsync, LinuxVolumeControlAsync)
        ):
            coro = volume_backend.fade(**kwargs)
            if blocking:
                await coro
            else:
                return loop.create_task(coro)

        else:
            loop.run_in_executor(None, partial(volume_backend.fade, **kwargs))
        return None

    @db_session
    async def play_recommended_tracks(
        self, time_range=TimeRange.LONG_TERM, device=None, **kwargs
    ):
        fade_args = kwargs.get("fade_args") or {
            k[5:]: v for k, v in kwargs.items() if k.startswith("fade_")
        }
        recommendation_args = kwargs.get("recommendation_args") or {
            k[4:]: v for k, v in kwargs.items() if k.startswith("rec_")
        }
        recommendation_args["time_range"] = time_range
        tracks = await self.recommend_by_top_artists(**recommendation_args)
        await self.fade_up(device=device, **fade_args)
        result = await self.start_playback(tracks=tracks, device=device)
        return {"playing": True, "device": device, "tracks": tracks, "result": result}

    @db_session
    async def play_recommended_genre(
        self, time_range=TimeRange.LONG_TERM, device=None, **kwargs
    ):
        fade_args = kwargs.get("fade_args") or {
            k[5:]: v for k, v in kwargs.items() if k.startswith("fade_")
        }
        popularity = random.choice(list(Playlist.Popularity)[:3])
        top_genres = await self.top_genres(time_range=time_range)
        genre = top_genres.select().without_distinct().random(1)[0]
        playlist = self.genre_playlist(genre.name, popularity)
        while not playlist:
            playlist = self.genre_playlist(genre.name, popularity)
        await self.fade_up(device=device, **fade_args)
        result = await playlist.play(self, device=device)
        return {
            "playing": True,
            "device": device,
            "playlist": playlist.to_dict(),
            "result": result,
        }

    @db_session
    async def play(
        self, time_range=TimeRange.LONG_TERM, device=None, item_type=None, **kwargs
    ):
        item_type = item_type or random.choice([ItemType.TRACKS, ItemType.PLAYLIST])
        if item_type == ItemType.TRACKS:
            return await self.play_recommended_tracks(time_range, device, **kwargs)

        if item_type == ItemType.PLAYLIST:
            return await self.play_recommended_genre(time_range, device, **kwargs)

        return {"playing": False}



import asyncio
import socket
import threading
import uuid
from datetime import datetime
from pathlib import Path

import addict
import aiohttp
import aiohttp.web
from aiohttp.web_runner import GracefulExit
from oauthlib.oauth2 import BackendApplicationClient
from pony.orm import db_session, get, select

from ... import config, logger, root
from ...cache import Country, User
from ...constants import API, AllScopes, AuthFlow
from ...exceptions import SpotifyCredentialsException
from ...sql import SQL
from .aiohttp_oauthlib import OAuth2Session

AUTH_HTML_FILE = root / "html" / "auth_message.html"
CACHE_FILE = Path.home() / ".cache" / "spfy" / ".web_cache"
web_app = aiohttp.web.Application()


def run_app(loop):
    asyncio.set_event_loop(loop)
    aiohttp.web.run_app(
        web_app, host="0.0.0.0", port=config.auth.callback.port, handle_signals=False
    )


class AuthMixin:
    def __init__(
        self,
        *args,
        client_id=None,
        client_secret=None,
        redirect_uri=None,
        user_id=None,
        username=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.client_id = client_id or config.app.client_id
        self.client_secret = client_secret or config.app.client_secret
        self.redirect_uri = self._get_redirect_uri(redirect_uri)
        self.user_id = user_id
        self.username = username
        self.callback_reached = threading.Event()
        self.flow = None
        self._session = None
        self.callback_loop = asyncio.new_event_loop()

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, new_session):
        if self._session and new_session != self._session:
            asyncio.ensure_future(self._session.close())
        self._session = new_session

    @staticmethod
    def _get_redirect_uri(redirect_uri):
        redirect_uri = (
            redirect_uri
            or config.app.redirect_uri
            or f"http://{socket.gethostname()}.local"
        )
        if config.auth.callback.enabled and config.auth.callback.port and redirect_uri:
            redirect_uri += f":{config.auth.callback.port}"
        return redirect_uri

    async def fetch_user(self, conn=None, op="OR", **fields):
        conn = conn or await self.dbpool

        if not (self.user_id or fields):
            return None

        user = None
        fields = {f: v for f, v in fields.items() if f}

        if self.user_id and not fields:
            user = await conn.fetchrow(SQL.user, self.user_id)
        else:
            condition = f" {op} ".join(
                f"{field} = ${i + 1}" for i, field in enumerate(fields.keys())
            )
            user = await conn.fetchrow(
                f"SELECT * FROM users WHERE {condition}", *fields.values()
            )

        if not user:
            return None

        return addict.Dict(dict(user))

    @property
    @db_session
    def user(self):
        return User[self.user_id]

    @property
    def is_authenticated(self):
        return bool(self.session and self.session.authorized)

    async def update_user_token(self, token):
        conn = await self.dbpool
        await conn.execute(SQL.update_user_token, token, self.user_id)

    # pylint: disable=too-many-locals
    async def authenticate_user_pg(
        self,
        username=None,
        email=None,
        code=None,
        state=None,
        auth_response=None,
        scope=AllScopes,
        conn=None,
    ):
        conn = conn or await self.dbpool

        self.flow = AuthFlow.AUTHORIZATION_CODE
        self.session = OAuth2Session(
            self.client_id,
            redirect_uri=self.redirect_uri,
            scope=scope,
            auto_refresh_url=API.TOKEN.value,
        )
        if self.user_id:
            user = await self.fetch_user()
            if user and user.token:
                self.session.token = user.token
                self.session.token_updater = self.update_user_token
                return user

        if username or email:
            user = await self.fetch_user(username=username, email=email)
            if user:
                self.user_id = user.id
                self.username = user.username
                self.session.token = user.token
                self.session.token_updater = self.update_user_token
                return user

        if code or auth_response:
            token = await self.session.fetch_token(
                token_url=API.TOKEN.value,
                client_id=self.client_id,
                client_secret=self.client_secret,
                code=code,
                state=state,
                authorization_response=auth_response,
            )
            self.session.token_updater = self.update_user_token

            user_details = await self.current_user()
            if user_details.birthdate:
                birthdate = datetime.strptime(user_details.birthdate, "%Y-%m-%d")
            else:
                birthdate = None

            iso_country = Country.get_iso_country(user_details.country)
            if not iso_country:
                country_name = country_code = user_details.country
            else:
                country_name = iso_country.name
                country_code = iso_country.alpha_2

            new_user_id = self.user_id or uuid.uuid4()
            user = await conn.fetchrow(
                SQL.upsert_user,
                new_user_id,
                user_details.email or str(new_user_id),
                user_details.id or "",
                user_details.country or "",
                user_details.display_name or "",
                birthdate,
                token,
                user_details.product == "premium",
                country_code,
                country_name,
            )
            user = addict.Dict(dict(user))
            self.user_id = user.id
            self.username = user.username

            if user_details.images:
                await conn.executemany(
                    """INSERT INTO images (
                            url, height, width, "user", unsplash_id,
                            unsplash_user_fullname, unsplash_user_username
                        ) VALUES ($1, $2, $3, $4, '', '', '')
                        ON CONFLICT DO NOTHING""",
                    [
                        (i.url, i.height or None, i.width or None, user.id)
                        for i in user_details.images
                        if i.url
                    ],
                )
            return user

        return None

    async def authenticate_server_pg(self, conn=None):
        conn = conn or await self.dbpool

        self.flow = AuthFlow.CLIENT_CREDENTIALS

        USA = Country.get_iso_country("US")
        default_user = await conn.fetchrow(
            SQL.upsert_user,
            User.DEFAULT_USERID,
            User.DEFAULT_EMAIL,
            User.DEFAULT_USERNAME,
            USA.alpha_2,
            User.DEFAULT_USERNAME,
            None,
            {},
            False,
            USA.alpha_2,
            USA.name,
        )
        default_user = addict.Dict(dict(default_user))

        self.user_id = default_user.id
        self.username = default_user.username
        self.session = OAuth2Session(client=BackendApplicationClient(self.client_id))
        self.session.token_updater = self.update_user_token
        if default_user.token:
            self.session.token = default_user.token
        else:
            default_user.token = await self.session.fetch_token(
                token_url=API.TOKEN.value,
                client_id=self.client_id,
                client_secret=self.client_secret,
            )
        return self.session

    async def authenticate_user(
        self,
        username=None,
        email=None,
        code=None,
        state=None,
        auth_response=None,
        scope=AllScopes,
    ):
        self.flow = AuthFlow.AUTHORIZATION_CODE
        self.session = OAuth2Session(
            self.client_id,
            redirect_uri=self.redirect_uri,
            scope=scope,
            auto_refresh_url=API.TOKEN.value,
        )
        with db_session:
            if self.user_id:
                user = User.get(id=self.user_id)
                if user and user.token:
                    self.session.token = user.token
                    self.session.token_updater = User.token_updater(user.id)
                    return self.session

            if username or email:
                if username:
                    user = get(u for u in User if u.username == username)
                elif email:
                    user = get(u for u in User if u.email == email)
                if user:
                    self.user_id = user.id
                    self.username = user.username
                    self.session.token = user.token
                    self.session.token_updater = User.token_updater(user.id)
                    return self.session

            if code or auth_response:
                token = await self.session.fetch_token(
                    token_url=API.TOKEN.value,
                    client_id=self.client_id,
                    client_secret=self.client_secret,
                    code=code,
                    state=state,
                    authorization_response=auth_response,
                )
                user_details = await self.current_user()
                user = (
                    select(
                        u
                        for u in User
                        if u.username == user_details.id
                        or u.email == user_details.email
                    )
                    .for_update()
                    .get()
                )
                if user:
                    user.token = token
                    if user.id != self.user_id:
                        self.user_id = user.id
                        self.username = user.username
                elif self.user_id:
                    user = User.get_for_update(id=self.user_id)
                    if user:
                        user.token = token
                if not user:
                    self.user_id = self.user_id or uuid.uuid4()
                    user_details["user_id"] = self.user_id
                    user_details["token"] = token
                    user = User.from_dict(user_details)
                    self.username = user.username
                return self.session

        return self.session

    @db_session
    async def authenticate_server(self):
        self.flow = AuthFlow.CLIENT_CREDENTIALS
        default_user = User.default()
        self.user_id = default_user.id
        self.username = default_user.username
        self.session = OAuth2Session(client=BackendApplicationClient(self.client_id))
        self.session.token_updater = User.token_updater(default_user.id)
        if default_user.token:
            self.session.token = default_user.token
        else:
            default_user.token = await self.session.fetch_token(
                token_url=API.TOKEN.value,
                client_id=self.client_id,
                client_secret=self.client_secret,
            )
        return self.session

    async def authenticate(self, flow=config.auth.flow, **auth_params):
        if not (self.client_id and self.client_secret):
            raise SpotifyCredentialsException

        self.flow = AuthFlow(flow)
        if self.flow == AuthFlow.CLIENT_CREDENTIALS:
            await self.authenticate_server()
        elif self.flow == AuthFlow.AUTHORIZATION_CODE:
            await self.authenticate_user(**auth_params)
            if not self.session.token:
                if config.auth.callback.enabled:
                    self.start_callback()
                authorization_url, _ = self.session.authorization_url(
                    API.AUTHORIZE.value
                )
                if config.auth.send_auth_url_to_email:
                    email = auth_params.get("email") or config.auth.email
                    self.send_auth_email(email, authorization_url)
                else:
                    print(f"Login here: {authorization_url}")
                await self.wait_for_authorization()

    async def wait_for_authorization(self):
        if not config.auth.callback.enabled:
            url = input("Paste the URL you are redirected to:")
            await self.authenticate_user(auth_response=url)
        else:
            self.callback_reached.wait(config.auth.callback.timeout)
            await self.stop_callback()

    @staticmethod
    async def stop_callback():
        try:
            async with aiohttp.request(
                "GET", f"http://localhost:{config.auth.callback.port}"
            ) as resp:
                await resp.read()
        except aiohttp.client_exceptions.ServerDisconnectedError:
            pass

    def start_callback(self):
        self.callback_reached.clear()

        async def callback(request):
            if self.callback_reached.is_set():
                raise GracefulExit

            code = request.query["code"]
            state = request.query.get("code")
            html = AUTH_HTML_FILE.read_text()  # pylint: disable=no-member
            try:
                asyncio.run_coroutine_threadsafe(
                    self.authenticate_user(code=code, state=state),
                    loop=self.callback_loop,
                ).result(config.auth.callback.timeout)
                html = html.replace("SPOTIFY_AUTH_MESSAGE", "Successfully logged in!")
                html = html.replace("BACKGROUND_COLOR", "#65D46E")
            except Exception as exc:
                logger.exception(exc)
                html = html.replace(
                    "SPOTIFY_AUTH_MESSAGE", "Could not get authorization token."
                )
                html = html.replace("BACKGROUND_COLOR", "#EC2E50")
            finally:
                self.callback_reached.set()
            return aiohttp.web.Response(body=html, content_type="text/html")

        web_app.router.add_get("/", callback)
        threading.Thread(
            target=run_app, daemon=True, args=(self.callback_loop,)
        ).start()



import asyncio
import random
import time
from collections import defaultdict
from datetime import date, datetime, timedelta
from functools import partial
from itertools import chain
from typing import Iterator

from pony.orm import db_session, get, select
from pony.orm.core import CacheIndexError
from unsplash.errors import UnsplashConnectionError, UnsplashError

from ... import logger
from ...asynch import LimitedAsCompletedError, limited_as_completed
from ...cache import Artist, City, Country, Genre, ImageMixin, Playlist, format_param
from ...constants import TimeRange
from ...sql import SQL
from ...util import normalize_features


class RecommenderMixin:
    USER_LIST = ("particledetector", "thesoundsofspotify")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def _get_upsert_genres_query(genres, conn):
        genre_names = ",\n".join(f"({format_param(genre)})" for genre in genres)
        return conn.execute(
            f"""INSERT INTO genres (name)
                VALUES {genre_names}
                ON CONFLICT DO NOTHING"""
        )

    @staticmethod
    def _get_upsert_countries_query(countries, conn):
        country_values = ",\n".join(
            f"({format_param(country.alpha_2)}, {format_param(country.name)})"
            for country in countries
            if country
        )
        return conn.execute(
            f"""INSERT INTO countries (code, name)
            VALUES {country_values}
            ON CONFLICT DO NOTHING"""
        )

    @staticmethod
    def _get_upsert_cities_query(cities, conn):
        city_values = ",\n".join(
            f"({format_param(city)}, {format_param(country)})"
            for city, country in cities
        )
        return conn.execute(
            f"""INSERT INTO cities (name, country)
                VALUES {city_values}
                ON CONFLICT DO NOTHING"""
        )

    @staticmethod
    def _get_upsert_playlists_queries(playlist_dicts, iso_countries):
        playlist_items = defaultdict(list)
        for p in playlist_dicts:
            if p.get("country"):
                if not iso_countries.get(p["country"]):
                    continue
                p["country"] = iso_countries[p["country"]].alpha_2
            playlist_items[tuple(p.keys())].append(p.values())

        queries = []
        for columns, value_list in playlist_items.items():
            column_string = ", ".join(columns)
            param_string = ",".join(
                f"({', '.join(format_param(p) for p in values)})"
                for values in value_list
            )
            queries.append(
                f"""INSERT INTO playlists AS p ({column_string})
                VALUES {param_string}
                ON CONFLICT DO NOTHING"""
            )

        return queries

    @staticmethod
    async def _get_unsplash_image_requests(conn, genres, countries, cities):
        image_requests = []
        if genres:
            image_genres = {
                r[0]
                for r in await conn.fetch(
                    "SELECT DISTINCT genre FROM images WHERE genre IS NOT NULL"
                )
            }
            new_genres = set(genres) - image_genres
            image_requests += [
                partial(Genre.get_image_fields, image_key=genre, genre=genre)
                for genre in new_genres
            ]

        if countries:
            image_countries = {
                r[0]
                for r in await conn.fetch(
                    "SELECT DISTINCT country FROM images WHERE country IS NOT NULL"
                )
            }
            country_dict = {c.alpha_2: c.name for c in countries.values() if c}
            new_countries = country_dict.keys() - image_countries
            image_requests += [
                partial(
                    Country.get_image_fields, image_key=country_dict[code], country=code
                )
                for code in new_countries
            ]
        if cities:
            image_cities = {
                r[0]
                for r in await conn.fetch(
                    "SELECT DISTINCT city FROM images WHERE city IS NOT NULL"
                )
            }
            city_country_mapping = dict(cities)
            new_cities = city_country_mapping.keys() - image_cities
            image_requests += [
                partial(
                    City.get_image_fields,
                    image_key=f"{city} {city_country_mapping[city]}",
                    city=city,
                )
                for city in new_cities
            ]

        return image_requests

    async def _upsert_remaining_images(self, reqs, initial_reqs):
        minutes = 62 - datetime.utcnow().minute
        delay = (minutes) * 60
        logger.info("Retrying image upsert in %d minutes", minutes)
        await asyncio.sleep(delay)
        await self._upsert_images(reqs, initial_reqs=initial_reqs)

    async def _upsert_images(self, reqs, conn=None, initial_reqs=None):
        conn = conn or await self.dbpool

        concurrency_limit = 100
        if not isinstance(reqs, Iterator):
            reqs_iterator = (fetch() for fetch in reqs)
        else:
            reqs_iterator = reqs

        try:
            async for resp in limited_as_completed(
                reqs_iterator,
                concurrency_limit,
                ignore_exceptions=(UnsplashError, UnsplashConnectionError),
            ):
                if not resp:
                    continue

                image_fields, updated_fields = resp
                await ImageMixin.upsert_unsplash_image(
                    conn, image_fields, **updated_fields
                )
        except LimitedAsCompletedError as exc:
            for future in exc.remaining_futures:
                future.cancel()

            await asyncio.sleep(2)
            for future in exc.remaining_futures:
                if future.done():
                    try:
                        _ = future.exception()
                    except:
                        pass

            remaining = len(list(reqs_iterator)) + concurrency_limit
            initial_reqs = initial_reqs or reqs

            if exc.original_exc:
                logger.exception(exc.original_exc)
            logger.warning(
                "Remaining images to fetch: %d/%d", remaining, len(initial_reqs)
            )

            asyncio.get_event_loop().create_task(
                self._upsert_remaining_images(
                    (fetch() for fetch in initial_reqs[-remaining:]), initial_reqs
                )
            )

    # pylint: disable=too-many-locals
    async def fetch_playlists_pg(self, conn=None):
        conn = conn or await self.dbpool

        existing_playlists = {
            p[0] for p in await conn.fetch("SELECT id FROM playlists")
        }
        user_ids_str = ",\n".join(f"('{user_id}')" for user_id in self.USER_LIST)
        await conn.execute(
            f"""
            INSERT INTO spotify_users (id)
            VALUES {user_ids_str}
            ON CONFLICT DO NOTHING
        """
        )

        playlist_dicts = []
        for user in self.USER_LIST:
            user_playlists = await self.user_playlists(user)
            async for playlist in user_playlists.iterall(ignore_exceptions=True):
                if not playlist:
                    continue

                logger.info("Got %s", playlist.name)
                if playlist.id not in existing_playlists:
                    playlist_dicts.append(Playlist.from_dict_pg(playlist))

                existing_playlists.add(playlist.id)

        genres = [p["genre"] for p in playlist_dicts if p.get("genre")]
        countries = [p["country"] for p in playlist_dicts if p.get("country")]
        iso_countries = {
            country: Country.get_iso_country(country) for country in countries
        }
        cities = [(p["city"], p["country"]) for p in playlist_dicts if p.get("city")]

        if genres:
            await self._get_upsert_genres_query(genres, conn)
        if countries:
            await self._get_upsert_countries_query(iso_countries.values(), conn)
        if cities:
            await self._get_upsert_cities_query(cities, conn)

        playlist_queries = self._get_upsert_playlists_queries(
            playlist_dicts, iso_countries
        )
        for query in playlist_queries:
            await conn.execute(query)

        image_requests = await self._get_unsplash_image_requests(
            conn, genres, iso_countries, cities
        )
        await self._upsert_images(image_requests, conn=conn)

    async def fetch_playlists(self):
        with db_session:
            fetched_ids = set(select(p.id for p in Playlist))
            for user in self.USER_LIST:
                user_playlists = await self.user_playlists(user)
                async for playlist in user_playlists.iterall(ignore_exceptions=True):
                    if not playlist:
                        continue

                    logger.info("Got %s", playlist.name)
                    if playlist.id not in fetched_ids:
                        Playlist.from_dict(playlist)
                    fetched_ids.add(playlist.id)

    async def fetch_user_top(self, time_range):
        with db_session:
            self.user.top_artists.clear()
            top_artists = await self.current_user_top_artists(
                limit=50, time_range=time_range
            )
            async for artist in top_artists.iterall():
                if self.is_disliked_artist(artist):
                    continue

                if Artist.exists(id=artist.id):
                    self.user.top_artists.add(Artist[artist.id])
                else:
                    try:
                        artist = await Artist.from_dict_async(artist)
                    except CacheIndexError:
                        artist = Artist[artist.id]
                    self.user.top_artists.add(artist)
            self.user.top_genres = (
                self.user.top_artists.genres.distinct().keys()
                - self.user.disliked_genres
            )
            if self.user.top_expires_at is None:
                self.user.top_expires_at = {}
            self.user.top_expires_at[TimeRange(time_range).value] = time.mktime(
                (date.today() + timedelta(days=1)).timetuple()
            )

    def genre_playlist(
        self, genre, popularity=Playlist.Popularity.SOUND
    ):  # pylint: disable=no-self-use
        with db_session:
            if not Playlist.exists():
                raise Exception('You have to call "fetch_playlists" first.')

            return Playlist.get(
                genre=genre, popularity=Playlist.Popularity(popularity).value
            )

    async def get_dislikes_for_filtering(self, conn=None):
        conn = conn or await self.dbpool

        dislikes = await conn.fetch(SQL.user_artist_genre_dislikes, self.user_id)

        disliked_artists = {row[0][:-3] for row in dislikes if row[0][-2:] == "AR"}
        disliked_genres = {row[0][:-3] for row in dislikes if row[0][-2:] == "GE"}
        return disliked_artists, disliked_genres

    async def top_artists_pg(
        self,
        time_range=TimeRange.SHORT_TERM,
        ignore=None,
        limit=None,
        conn=None,
        dislikes=None,
    ):
        conn = conn or await self.dbpool

        disliked_artists, disliked_genres = (
            dislikes or await self.get_dislikes_for_filtering(conn)
        )
        top_artists = await self.current_user_top_artists(
            limit=50, time_range=time_range
        )
        ignore = set(ignore or [])
        top_artists = [
            artist
            for artist in top_artists
            if not self.is_disliked_artist(artist, disliked_artists, disliked_genres)
            and artist.id not in ignore
        ]
        if limit:
            top_artists = random.sample(top_artists, min(limit, len(top_artists)))

        return top_artists

    async def top_genres_pg(
        self,
        time_range=TimeRange.SHORT_TERM,
        ignore=None,
        limit=None,
        conn=None,
        dislikes=None,
    ):
        conn = conn or await self.dbpool

        artists = await self.top_artists_pg(
            time_range=time_range, conn=conn, dislikes=dislikes
        )
        genres = set(chain.from_iterable(artist.genres for artist in artists))

        if ignore:
            genres -= set(ignore)
        if limit:
            genres = random.sample(genres, min(limit, len(genres)))

        return list(genres)

    async def top_artists(self, time_range=TimeRange.SHORT_TERM):
        with db_session:
            if self.user.top_expired(time_range):
                await self.fetch_user_top(time_range)
            return self.user.top_artists

    async def top_genres(self, time_range=TimeRange.SHORT_TERM):
        with db_session:
            if self.user.top_expired(time_range):
                await self.fetch_user_top(time_range)
            return self.user.top_genres

    async def order_by(self, features, tracks):
        if tracks and isinstance(tracks[0], str):
            tracks = list(set(tracks))

        audio_features, tracks = await asyncio.gather(
            self.audio_features(tracks=tracks), self.tracks(tracks=tracks)
        )

        audio_features_by_id = {a.id: a.to_dict() for a in audio_features if a}
        for track in tracks:
            track_features = audio_features_by_id.get(track.id)
            if track_features and track_features.get("popularity") is None:
                track_features["popularity"] = track.popularity or 0

        track_ids = list(audio_features_by_id.keys())
        audio_features = [
            {feature: a.get(feature) or 0 for feature in features.keys()}
            for a in audio_features_by_id.values()
        ]
        audio_features = normalize_features(audio_features, track_ids)
        for feature, direction in features.items():
            audio_features[feature] *= direction
        audio_features["total"] = audio_features.sum(axis=1)
        return audio_features.sort_values("total").index.tolist()

    async def fill_with_related_artists(self, artists, limit=5):
        tries = 5
        if len(artists) >= limit:
            return artists

        artist_set = set(artists)
        artist_list = list(artists)
        while len(artist_set) < limit and tries:
            tries -= 1
            related_artists = {
                a.id
                for a in (await self.artist_related_artists(random.choice(artist_list)))
                if self.is_not_disliked_artist(a)
            }
            related_artists_limit = min(
                random.randint(1, limit - len(artist_set)), len(related_artists)
            )
            artist_set |= set(random.sample(related_artists, related_artists_limit))
        return artist_set

    async def recommend_by_top_artists(
        self,
        artist_limit=2,
        track_limit=100,
        use_related=True,
        features_order=None,
        time_range=TimeRange.SHORT_TERM,
        **kwargs,
    ):
        """Get a list of recommended songs.

        Returns:
            list: List of tracks
        """
        with db_session:
            top_artists = await self.top_artists(time_range=time_range)
            artists = top_artists.select().without_distinct().random(artist_limit)
            if use_related:
                artists = await self.fill_with_related_artists([a.id for a in artists])
            tracks = await self.recommendations(
                seed_artists=artists, limit=track_limit, **kwargs
            )
            tracks = list(filter(self.is_not_disliked_track, tracks))
            if features_order:
                tracks = await self.order_by(features_order, tracks)
            return tracks

    def is_disliked_artist(self, artist, disliked_artists=None, disliked_genres=None):
        if disliked_artists is not None and disliked_genres is not None:
            return artist.id in disliked_artists or bool(
                set(artist.genres or []) & disliked_genres
            )
        with db_session:
            return artist.id in set(
                self.user.disliked_artists.id.distinct().keys()
            ) or bool(set(artist.genres or []) & set(self.user.disliked_genres))

    def is_not_disliked_artist(self, artist):
        with db_session:
            return not self.is_disliked_artist(artist)

    def is_not_disliked_track(self, track):
        with db_session:
            return all(self.is_not_disliked_artist(a) for a in track.artists)

    def disliked_artists(self):
        with db_session:
            return list(self.user.disliked_artists)

    def disliked_genres(self):
        with db_session:
            return list(self.user.disliked_genres)

    async def dislike_artist(self, artist):
        with db_session:
            if isinstance(artist, str):
                # pylint: disable=consider-using-in
                artist = get(a for a in Artist if a.id == artist or a.name == artist)
            elif isinstance(artist, dict):
                artist = Artist.get(artist.id) or (await Artist.from_dict_async(artist))
            self.user.dislike(artist=artist)

    def dislike_genre(self, genre):
        with db_session:
            if isinstance(genre, str):
                genre = genre.lower()
                genre = Genre.get(name=genre) or Genre(name=genre)
            self.user.dislike(genre=genre)



# pylint: disable=too-many-lines
import asyncio
import os
import random
import re
import string
import time
from collections import OrderedDict
from datetime import date, datetime
from enum import IntEnum
from io import BytesIO
from uuid import NAMESPACE_URL, UUID, uuid4, uuid5

import aiohttp
import psycopg2.extras
import requests
from colorthief import ColorThief
from first import first
from pony.orm import (
    Database,
    Json,
    ObjectNotFound,
    Optional,
    PrimaryKey,
    Required,
    Set,
    db_session,
    desc,
    ormtypes,
    select,
    sql_debug,
)
from pony.orm.core import CacheIndexError
from psycopg2.extensions import register_adapter
from pycountry import countries
from unsplash.errors import UnsplashConnectionError, UnsplashError

from .. import Unsplash, config, logger
from ..constants import TimeRange
from ..sql import SQL, SQL_DEFAULT

DIGITS_PATTERN = re.compile(r"[0-9]")
register_adapter(ormtypes.TrackedDict, psycopg2.extras.Json)
if os.getenv("SQL_DEBUG"):
    sql_debug(True)
    import logging

    logging.getLogger("pony.orm.sql").setLevel(logging.DEBUG)
db = Database()


def create_condition(op="AND", firstsub=1, **fields):
    condition = f" {op} ".join(
        f"{field} = ${i + firstsub}" for i, field in enumerate(fields.keys())
    )
    return condition


def format_param(param):
    if isinstance(param, (int, float, bool)):
        return str(param)
    if isinstance(param, (str, bytes, dict, date, datetime, UUID)):
        tag = "".join(random.sample(string.ascii_letters, 3))
        return f"${tag}${param}${tag}$"
    if param is None:
        return "null"
    return param


class ImageMixin:
    @classmethod
    async def image_pg(cls, conn, width=None, height=None, **fields):
        condition = create_condition(firstsub=2, **fields)
        if width:
            image = await conn.fetchrow(
                f"SELECT * FROM images WHERE width >= $1 AND {condition} ORDER BY width LIMIT 1",
                width,
                *fields.values(),
            )
        elif height:
            image = await conn.fetchrow(
                f"SELECT * FROM images WHERE height >= $1 AND {condition} ORDER BY height LIMIT 1",
                height,
                *fields.values(),
            )
        else:
            image = await conn.fetchrow(
                f"SELECT * FROM images WHERE {condition} ORDER BY width DESC LIMIT 1",
                *fields.values(),
            )
        return image

    def image(self, width=None, height=None):
        if width:
            image = (
                self.images.select()
                .where(lambda i: i.width >= width)
                .order_by(Image.width)
                .first()
            )
        elif height:
            image = (
                self.images.select()
                .where(lambda i: i.height >= height)
                .order_by(Image.height)
                .first()
            )
        else:
            image = self.images.select().order_by(desc(Image.width)).first()
        return image

    @classmethod
    def get_image_queries_pg(cls, key):
        words = key.split()
        stems = [[w[:i] for i in range(len(w), 2, -1)] for w in words]
        queries = [key, *words, *sum(stems, [])]
        return queries + [f"{query} music" for query in queries]

    def get_image_queries(self):
        words = self.name.split()
        stems = [[w[:i] for i in range(len(w), 2, -1)] for w in words]
        queries = [*words, self.name, *sum(stems, [])]
        return [f"{query} music" for query in queries]

    @classmethod
    async def get_image_fields(cls, image_key=None, **fields):
        logger.info("Getting image for key=%s %s", image_key, fields)

        key = image_key or " ".join(str(v) for v in fields.values())
        queries = cls.get_image_queries_pg(key)

        photo = await cls.get_unsplash_photo(queries)
        if photo is None:
            return None

        ratio = photo.height / photo.width
        params = {
            "color": photo.color,
            "unsplash_id": photo.id,
            "unsplash_user_fullname": photo.user.name,
            "unsplash_user_username": photo.user.username,
            **fields,
        }
        image_fields = [
            {
                **params,
                "url": photo.urls.full,
                "width": photo.width,
                "height": photo.height,
            },
            {
                **params,
                "url": photo.urls.regular,
                "width": Image.REGULAR,
                "height": int(round(ratio * Image.REGULAR)),
            },
            {
                **params,
                "url": photo.urls.small,
                "width": Image.SMALL,
                "height": int(round(ratio * Image.SMALL)),
            },
            {
                **params,
                "url": photo.urls.thumb,
                "width": Image.THUMB,
                "height": int(round(ratio * Image.THUMB)),
            },
        ]
        image_fields = [
            OrderedDict(sorted(d.items(), key=lambda t: t[0])) for d in image_fields
        ]

        return image_fields, fields

    @classmethod
    async def upsert_unsplash_image(cls, conn, image_fields, **updated_fields):
        logger.debug(
            "Upserting image: %s | UPDATED: [%s]", image_fields, updated_fields
        )

        columns = ", ".join(image_fields[0].keys())
        values = ",\n".join(
            f"({', '.join([format_param(p) for p in im.values()])})"
            for im in image_fields
        )
        updated_fields_str = ", ".join(
            f"{col} = ${i+1}" for i, col in enumerate(updated_fields.keys())
        )
        images = await conn.fetch(
            f"""INSERT INTO images AS im ({columns})
            VALUES {values}
            ON CONFLICT (url) DO UPDATE SET {updated_fields_str}
            RETURNING *
            """,
            *updated_fields.values(),
        )
        return [dict(i) for i in images]

    @classmethod
    async def fetch_unsplash_image_pg(
        cls, conn, width=None, height=None, image_key=None, **fields
    ):
        image_fields, updated_fields = await cls.get_image_fields(image_key, **fields)
        images = await cls.upsert_unsplash_image(conn, image_fields, **updated_fields)

        if not images:
            return None
        return cls.get_optimal_image(images, width=width, height=height)

    @staticmethod
    def get_optimal_image(images, width=None, height=None):
        images = sorted(images, key=lambda i: i.get("width" or 0))
        if width:
            return first(
                images, key=lambda i: (i.get("width") or 0) >= width, default=images[0]
            )
        if height:
            return first(
                images,
                key=lambda i: (i.get("height") or 0) >= height,
                default=images[0],
            )
        return images[0]

    @staticmethod
    async def get_unsplash_photo(queries):
        photo = None
        for query in queries:
            try:
                photos = await Unsplash.photo.random(
                    query=query, orientation="squarish"
                )
            except (UnsplashError, UnsplashConnectionError):
                continue

            if photos:
                photo = photos[0]
                break

        else:
            try:
                photo = (
                    await Unsplash.photo.random(query="music", orientation="squarish")
                )[0]
            except (UnsplashError, UnsplashConnectionError):
                photo = None
        return photo

    async def fetch_unsplash_image(self, width=None, height=None):
        image = self.image(width, height)
        if image:
            return image

        queries = self.get_image_queries()
        photo = await self.get_unsplash_photo(queries)
        if photo is None:
            return None

        images = select(i for i in Image if i.unsplash_id == photo.id).for_update()
        if images.exists():
            params = {self.__class__.__name__.lower(): self}
            for image in images:
                image.set(**params)
            return self.image(width, height)

        params = {self.__class__.__name__.lower(): self, "unsplash_id": photo.id}
        image_exists = False
        for url in (
            photo.urls.full,
            photo.urls.regular,
            photo.urls.small,
            photo.urls.thumb,
        ):
            image = Image.get(url=url)
            if image:
                image_exists = True
                image.set(**params)
        if image_exists:
            return self.image(width, height)

        ratio = photo.height / photo.width
        params = {
            self.__class__.__name__.lower(): self,
            "color": photo.color,
            "unsplash_id": photo.id,
            "unsplash_user_fullname": photo.user.name,
            "unsplash_user_username": photo.user.username,
        }
        Image(url=photo.urls.full, width=photo.width, height=photo.height, **params)
        Image(
            url=photo.urls.regular,
            width=Image.REGULAR,
            height=int(round(ratio * Image.REGULAR)),
            **params,
        )
        Image(
            url=photo.urls.small,
            width=Image.SMALL,
            height=int(round(ratio * Image.SMALL)),
            **params,
        )
        Image(
            url=photo.urls.thumb,
            width=Image.THUMB,
            height=int(round(ratio * Image.THUMB)),
            **params,
        )
        return self.image(width, height)


class User(db.Entity, ImageMixin):
    _table_ = "users"
    DEFAULT_EMAIL = "spfy@backend"
    DEFAULT_USERNAME = "spfy-backend"
    DEFAULT_USERID = uuid5(NAMESPACE_URL, DEFAULT_USERNAME)
    id = PrimaryKey(
        UUID, default=uuid4, sql_default=SQL_DEFAULT.uuid4
    )  # pylint: disable=redefined-builtin
    email = Optional(str, unique=True, index=True, sql_default="''")
    username = Required(str, unique=True, index=True)
    country = Required("Country")
    preferred_country = Optional("Country")
    spotify_user = Optional("SpotifyUser")
    display_name = Optional(str, sql_default="''")
    birthdate = Optional(date)
    token = Required(Json, volatile=True)
    spotify_premium = Required(bool)
    api_calls = Required(int, default=0, sql_default="0", volatile=True)
    created_at = Required(
        datetime, default=datetime.utcnow, sql_default=SQL_DEFAULT.now
    )
    last_usage_at = Required(
        datetime, default=datetime.utcnow, sql_default=SQL_DEFAULT.now, volatile=True
    )
    images = Set("Image", cascade_delete=True)
    top_artists = Set("Artist")
    disliked_artists = Set("Artist")
    top_genres = Set("Genre")
    disliked_genres = Set("Genre")
    top_countries = Set("Country")
    disliked_countries = Set("Country")
    top_cities = Set("City")
    disliked_cities = Set("City")
    top_expires_at = Required(Json, volatile=True, default=dict, sql_default="'{}'")

    def to_dict(self, *args, **kwargs):  # pylint: disable=arguments-differ
        _dict = super().to_dict(*args, **kwargs)
        if "id" in _dict:
            _dict["id"] = str(_dict["id"])
        return _dict

    @classmethod
    async def from_dict_async(cls, user):
        if user.images:
            try:
                color = await Image.grab_color_async(user.images[-1].url)
            except:
                color = "#000000"
            for image in user.images:
                image.color = color
        return cls.from_dict(user, grab_image_color=False)

    @classmethod
    def from_dict(cls, user, grab_image_color=True):
        if user.images and grab_image_color:
            try:
                color = Image.grab_color(user.images[-1].url)
            except:
                color = "#000000"
            for image in user.images:
                image.color = color
        images = [Image.get(url=image.url) or Image(**image) for image in user.images]
        spotify_user = SpotifyUser.get(id=user.id) or SpotifyUser(
            id=user.id, name=user.get("display_name") or ""
        )
        return cls(
            id=user.user_id,
            spotify_premium=user.product == "premium",
            spotify_user=spotify_user,
            username=user.id,
            email=user.email or str(user.user_id),
            token=user.token,
            country=Country.from_str(user.country),
            images=images,
            display_name=user.display_name or "",
            birthdate=datetime.strptime(user.birthdate, "%Y-%m-%d")
            if user.birthdate
            else None,
        )

    async def _fetch_artist(self, artist, client):
        if artist and client:
            if Artist.exists(id=artist):
                return artist

            artist_result = await client.artist(artist)
            if not artist_result:
                return artist

            try:
                with db_session:
                    # pylint: disable=unused-variable
                    artist_row = await Artist.from_dict_async(artist_result)
            except CacheIndexError:
                pass
        return artist

    async def dislike_async(
        self, artist=None, genre=None, country=None, city=None, client=None
    ):
        artist = await self._fetch_artist(artist, client)
        self.dislike(artist=artist, genre=genre, country=country, city=city)

    async def like_async(
        self, artist=None, genre=None, country=None, city=None, client=None
    ):
        artist = await self._fetch_artist(artist, client)
        self.like(artist=artist, genre=genre, country=country, city=city)

    def dislike(self, artist=None, genre=None, country=None, city=None):
        assert artist or genre or country or city
        if artist:
            artist = Artist.get(id=artist)
            if artist:
                self.disliked_artists.add(artist)
                self.top_artists.remove(artist)
        if genre:
            genre = Genre.get(name=genre) or Genre(name=genre)
            if genre:
                self.disliked_genres.add(genre)
                self.top_genres.remove(genre)
                self.disliked_artists.add(genre.artists)
                self.top_artists.remove(genre.artists)
        if country:
            country = Country.get(code=country)
            if country:
                self.disliked_countries.add(country)
                self.top_countries.remove(country)
        if city:
            city = City.get(name=city)
            if city:
                self.disliked_cities.add(city)
                self.top_cities.remove(city)

    def like(self, artist=None, genre=None, country=None, city=None):
        assert artist or genre or country or city
        if artist:
            artist = Artist.get(id=artist)
            if artist:
                self.top_artists.add(artist)
                self.disliked_artists.remove(artist)
        if genre:
            genre = Genre.get(name=genre)
            if genre:
                self.top_genres.add(genre)
                self.disliked_genres.remove(genre)
                self.top_artists.add(genre.artists)
                self.disliked_artists.remove(genre.artists)
        if country:
            country = Country.get(code=country)
            if country:
                self.top_countries.add(country)
                self.disliked_countries.remove(country)
        if city:
            city = City.get(name=city)
            if city:
                self.top_cities.add(city)
                self.disliked_cities.remove(city)

    @classmethod
    async def like_pg(cls, conn, user_id, **kwargs):
        for key, value in kwargs.items():
            await conn.execute(SQL.like.format(key), user_id, value)

    @classmethod
    async def dislike_pg(cls, conn, spotify, **kwargs):
        artist = kwargs.pop("artist", None)
        if artist:
            artist = await spotify.artist(artist)
            await conn.execute(
                SQL.dislike_artist.format("artist"),
                spotify.user_id,
                artist.id,
                artist.name,
                artist.followers.total or 0,
                artist.popularity or 0,
            )
        for key, value in kwargs.items():
            await conn.execute(SQL.dislike.format(key), spotify.user_id, value)

    def top_expired(self, time_range):
        time_range = TimeRange(time_range).value
        return (
            not self.top_expires_at
            or time_range not in self.top_expires_at
            or time.time() >= self.top_expires_at[time_range]
        )

    @classmethod
    def default(cls):
        try:
            return User[cls.DEFAULT_USERID]

        except ObjectNotFound:
            return User(
                id=cls.DEFAULT_USERID,
                username=cls.DEFAULT_USERNAME,
                email=cls.DEFAULT_EMAIL,
                token={},
                country=Country.from_str("US"),
                spotify_premium=False,
            )

    @staticmethod
    def token_updater(_id):
        @db_session
        def update(token):
            User[_id].token = token

        return update


class Image(db.Entity):
    _table_ = "images"
    REGULAR = 1080
    SMALL = 400
    THUMB = 200
    url = PrimaryKey(str)
    height = Optional(int)
    width = Optional(int)
    color = Optional(str, sql_default="''")
    playlist = Optional("Playlist")
    artist = Optional("Artist")
    genre = Optional("Genre")
    country = Optional("Country")
    city = Optional("City")
    user = Optional("User")
    unsplash_id = Optional(str, index=True, sql_default="''")
    unsplash_user_fullname = Optional(str, sql_default="''")
    unsplash_user_username = Optional(str, sql_default="''")

    # pylint: disable=no-self-use

    @classmethod
    def unsplash_url(cls):
        return f"https://unsplash.com/?utm_source={config.unsplash.app_name}&utm_medium=referral"

    @classmethod
    def unsplash_user_url(cls, username):
        return f"https://unsplash.com/@{username}?utm_source={config.unsplash.app_name}&utm_medium=referral"

    @classmethod
    def unsplash_credits(cls, user_fullname, username):
        return {
            "user_name": user_fullname,
            "user_url": cls.unsplash_user_url(username),
            "site_url": cls.unsplash_url(),
        }

    @staticmethod
    async def grab_color_async(image_url):
        async with aiohttp.ClientSession() as client:
            async with client.get(image_url) as resp:
                image_file = BytesIO(await resp.read())
                color = ColorThief(image_file).get_color(quality=1)
                return f"#{color[0]:02x}{color[1]:02x}{color[2]:02x}"

    @staticmethod
    def grab_color(image_url):
        resp = requests.get(image_url)
        image_file = BytesIO(resp.content)
        color = ColorThief(image_file).get_color(quality=1)
        return f"#{color[0]:02x}{color[1]:02x}{color[2]:02x}"

    @classmethod
    async def download_pg(cls, conn, url):
        unsplash_id = await conn.fetchval(
            "SELECT unsplash_id FROM images WHERE url = $1", url
        )
        image = None
        async with aiohttp.ClientSession() as client:
            async with client.get(url) as resp:
                _, image = await asyncio.gather(
                    Unsplash.photo.download(unsplash_id, without_content=True),
                    resp.read(),
                )
        return image

    async def download(self):
        image = None
        async with aiohttp.ClientSession() as client:
            async with client.get(self.url) as resp:
                _, image = await asyncio.gather(
                    Unsplash.photo.download(self.unsplash_id, without_content=True),
                    resp.read(),
                )
        return image


class SpotifyUser(db.Entity):
    _table_ = "spotify_users"
    id = PrimaryKey(str)  # pylint: disable=redefined-builtin
    name = Optional(str, sql_default="''")
    user = Optional("User")
    playlists = Set("Playlist")

    @property
    def uri(self):
        return f"spotify:user:{self.id}"

    @property
    def href(self):
        return f"https://api.spotify.com/v1/users/{self.id}"

    @property
    def external_url(self):
        return f"http://open.spotify.com/user/{self.id}"


class Genre(db.Entity, ImageMixin):
    _table_ = "genres"
    name = PrimaryKey(str)
    artists = Set("Artist")
    playlists = Set("Playlist")
    fans = Set(User, reverse="top_genres", table="genre_fans")
    haters = Set(User, reverse="disliked_genres", table="genre_haters")
    images = Set(Image, cascade_delete=True)

    def play(self, client, device=None):
        popularity = random.choice(list(Playlist.Popularity)[:3])
        playlist = client.genre_playlist(self.name, popularity)
        return playlist.play(device=device)


class Country(db.Entity, ImageMixin):
    _table_ = "countries"
    code = PrimaryKey(str, max_len=2)
    name = Required(str, index=True)
    users = Set("User", reverse="country")
    users_preferring = Set("User", reverse="preferred_country")
    cities = Set("City")
    playlists = Set("Playlist")
    fans = Set(User, reverse="top_countries", table="country_fans")
    haters = Set(User, reverse="disliked_countries", table="country_haters")
    images = Set(Image, cascade_delete=True)

    @staticmethod
    def _name_match(country, name):
        try:
            return name in country.name.lower() or name in country.official_name.lower()

        except:
            return False

    @classmethod
    def get_iso_country(cls, country):
        iso_country, code, name = None, None, None
        if len(country) == 2:
            code = country
        else:
            name = country

        if country == "UK":
            iso_country = countries.get(alpha_2="GB")
        elif country == "USA":
            iso_country = countries.get(alpha_2="US")
        elif code is not None:
            try:
                iso_country = countries.get(alpha_2=code)
            except KeyError:
                iso_country = first(
                    countries, key=lambda c: code.lower() in c.alpha_2.lower()
                )
        elif name is not None:
            try:
                iso_country = countries.get(name=name)
            except KeyError:
                try:
                    iso_country = countries.get(official_name=name)
                except KeyError:
                    try:
                        iso_country = countries.get(common_name=name)
                    except KeyError:
                        iso_country = first(
                            countries, key=lambda c: cls._name_match(c, name.lower())
                        )
        if not iso_country:
            logger.error(
                "Could not find a country with name=%s and code=%s", name, code
            )
            return None
        return iso_country

    @classmethod
    def from_str(cls, country):
        iso_country = cls.get_iso_country(country)
        if not iso_country:
            return None

        return cls.get(name=iso_country.name) or cls(
            name=iso_country.name, code=iso_country.alpha_2
        )

    def get_image_queries(self):
        words = self.name.split()
        stems = [[w[:i] for i in range(len(w), 2, -1)] for w in words]
        queries = [self.name, *words, *sum(stems, [])]
        return queries


class City(db.Entity, ImageMixin):
    _table_ = "cities"
    name = PrimaryKey(str)
    country = Required(Country)
    playlists = Set("Playlist")
    fans = Set(User, reverse="top_cities", table="city_fans")
    haters = Set(User, reverse="disliked_cities", table="city_haters")
    images = Set(Image, cascade_delete=True)

    def get_image_queries(self):
        words = self.name.split()
        stems = [[w[:i] for i in range(len(w), 2, -1)] for w in words]
        queries = [self.name, self.country.name, *words, *sum(stems, [])]
        return queries


class Playlist(db.Entity, ImageMixin):
    _table_ = "playlists"
    YEAR = "(?P<year>[0-9]{4})"
    GENRE = "(?P<genre>.+)"
    CITY = "(?P<city>.+)"
    COUNTRY = "(?P<country>.+?)"
    COUNTRY_CODE = "(?P<country_code>[A-Z]{2})"
    DATE = "(?P<date>[0-9]{8})"
    INTRO_POPULARITY = "(?P<popularity>Intro)"
    GENRE_POPULARITY_TITLE = "(?P<popularity>Sound|Pulse|Edge)"
    GENRE_POPULARITY_LOWER = "(?P<popularity>sound|pulse|edge)"
    NEEDLE_POPULARITY = "(?P<popularity>Current|Emerging|Underground)"
    PATTERNS = OrderedDict(
        intro_to_genre=re.compile(f"^{INTRO_POPULARITY} to {GENRE}$"),
        sound_of_city=re.compile(f"^The Sound of {CITY} {COUNTRY_CODE}$"),
        needle=re.compile(f"^The Needle / {COUNTRY}(?: - {NEEDLE_POPULARITY})?$"),
        pine_needle=re.compile(f"^The Pine Needle / {COUNTRY}$"),
        year_in_genre=re.compile(f"^{YEAR} in {GENRE}$"),
        meta_genre=re.compile(f"^Meta{GENRE_POPULARITY_LOWER}: {GENRE}$"),
        meta_year_in_genre=re.compile(f"^Meta{YEAR}: {GENRE}$"),
        sound_of_genre=re.compile(f"^The {GENRE_POPULARITY_TITLE} of {GENRE}$"),
        women_filter_genre=re.compile(f"^A ♀Filter for {GENRE}$"),
    )

    class Popularity(IntEnum):
        SOUND = 0
        PULSE = 1
        EDGE = 2
        CURRENT = 3
        EMERGING = 4
        UNDERGROUND = 5
        YEAR = 6
        ALL = 7
        INTRO = 8

    id = PrimaryKey(str)  # pylint: disable=redefined-builtin
    collaborative = Required(bool)
    name = Required(str)
    description = Optional(str, sql_default="''")
    owner = Required(SpotifyUser)
    public = Required(bool)
    snapshot_id = Required(str)
    tracks = Required(int)
    popularity = Optional(int, index=True, sql_default="0")
    genre = Optional(Genre)
    country = Optional(Country)
    city = Optional(City)
    date = Optional(date, index=True)
    year = Optional(int)
    women = Optional(bool, index=True, sql_default=SQL_DEFAULT.bool_false)
    christmas = Optional(bool, index=True, sql_default=SQL_DEFAULT.bool_false)
    meta = Optional(bool, index=True, sql_default=SQL_DEFAULT.bool_false)
    images = Set(Image, cascade_delete=True)

    def play(self, client, device=None):
        return client.start_playback(playlist=self.uri, device=device)

    @property
    def uri(self):
        return f"spotify:user:{self.owner.id}:playlist:{self.id}"

    @property
    def href(self):
        return f"https://api.spotify.com/v1/users/{self.owner.id}/playlists/{self.id}"

    @property
    def external_url(self):
        return f"http://open.spotify.com/user/{self.owner.id}/playlist/{self.id}"

    @classmethod
    def get_fields(cls, groups):
        fields = {}
        if "year" in groups:
            fields["year"] = int(groups["year"])
            fields["date"] = date(fields["year"], 1, 1)
            fields["popularity"] = cls.Popularity.YEAR.value
        if "city" in groups and "country_code" in groups:
            fields["country"] = groups["country_code"]
            fields["city"] = groups["city"]
            fields["popularity"] = cls.Popularity.SOUND.value
        if "genre" in groups:
            genre = groups["genre"].lower()
            fields["genre"] = genre
            if "christmas" in genre:
                fields["christmas"] = True
        if "country" in groups:
            fields["country"] = groups["country"]
        if groups.get("date"):
            fields["date"] = datetime.strptime(groups["date"], "%Y%m%d").date()
        if "popularity" in groups:
            popularity = groups["popularity"]
            if popularity:
                fields["popularity"] = cls.Popularity[popularity.upper()].value
            else:
                fields["popularity"] = cls.Popularity.ALL.value
        return fields

    @classmethod
    def from_dict_pg(cls, playlist):
        fields = {
            "id": playlist.id,
            "description": playlist.description or "",
            "collaborative": playlist.collaborative,
            "name": playlist.name,
            "owner": playlist.owner.id,
            "public": playlist.public,
            "snapshot_id": playlist.snapshot_id,
            "tracks": playlist.tracks.total,
            "women": playlist.name.startswith("A ♀Filter for"),
            "christmas": (
                "Pine Needle" in playlist.name or "christmas" in playlist.name.lower()
            ),
            "meta": playlist.name.startswith("Meta"),
        }
        for pattern in cls.PATTERNS.values():
            match = pattern.match(playlist.name)
            if match:
                groups = match.groupdict()
                fields.update(cls.get_fields(groups))
                break

        else:
            logger.warning("No pattern matches the playlist: %s", playlist.name)
        return fields

    @classmethod
    def from_dict(cls, playlist):
        owner = SpotifyUser.get(id=playlist.owner.id) or SpotifyUser(
            id=playlist.owner.id, name=playlist.owner.get("display_name") or ""
        )
        fields = {
            "id": playlist.id,
            "collaborative": playlist.collaborative,
            "name": playlist.name,
            "owner": owner,
            "public": playlist.public,
            "snapshot_id": playlist.snapshot_id,
            "tracks": playlist.tracks.total,
            "christmas": (
                "Pine Needle" in playlist.name or "christmas" in playlist.name.lower()
            ),
            "meta": playlist.name.startswith("Meta"),
            "images": [Image.get(url=im.url) or Image(**im) for im in playlist.images],
        }
        for pattern in cls.PATTERNS.values():
            match = pattern.match(playlist.name)
            if match:
                groups = match.groupdict()
                fields.update(cls.get_fields(groups))
                break

        else:
            logger.warning("No pattern matches the playlist: %s", playlist.name)

        if fields.get("country"):
            fields["country"] = Country.from_str(fields["country"])
        if fields.get("city"):
            fields["city"] = City.get(name=fields["city"]) or City(
                name=fields["city"], country=fields["country"]
            )
        if fields.get("genre"):
            fields["genre"] = Genre.get(name=fields["genre"]) or Genre(
                name=fields["genre"]
            )
        return cls(**fields)


class Artist(db.Entity, ImageMixin):
    _table_ = "artists"
    id = PrimaryKey(str)  # pylint: disable=redefined-builtin
    name = Required(str, index=True)
    followers = Required(int)
    genres = Set(Genre)
    fans = Set(User, reverse="top_artists", table="artist_fans")
    haters = Set(User, reverse="disliked_artists", table="artist_haters")
    popularity = Optional(int, sql_default="0")
    images = Set(Image, cascade_delete=True)

    def play(self, client, device=None):
        return client.start_playback(artist=self.uri, device=device)

    @property
    def uri(self):
        return f"spotify:artist:{self.id}"

    @property
    def href(self):
        return f"https://api.spotify.com/v1/artists/{self.id}"

    @property
    def external_url(self):
        return f"http://open.spotify.com/artist/{self.id}"

    @classmethod
    async def from_dict_async(cls, artist):
        if artist.images:
            try:
                color = await Image.grab_color_async(artist.images[-1].url)
            except:
                color = "#000000"
            for image in artist.images:
                image.color = color
        return cls.from_dict(artist, grab_image_color=False)

    @classmethod
    def from_dict(cls, artist, grab_image_color=True):
        if artist.images and grab_image_color:
            try:
                color = Image.grab_color(artist.images[-1].url)
            except:
                color = "#000000"
            for image in artist.images:
                image.color = color
        genres = [Genre.get(name=genre) or Genre(name=genre) for genre in artist.genres]
        images = [Image.get(url=image.url) or Image(**image) for image in artist.images]
        return cls(
            id=artist.id,
            name=artist.name,
            followers=artist.followers.total,
            genres=genres,
            images=images,
            popularity=artist.popularity,
        )


# pylint: disable=too-few-public-methods


class AudioFeatures(db.Entity):
    _table_ = "audio_features"
    KEYS = ["C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"]
    id = PrimaryKey(str)  # pylint: disable=redefined-builtin
    acousticness = Required(float, min=0.0, max=1.0)
    danceability = Required(float, min=0.0, max=1.0)
    duration_ms = Required(int, min=0)
    energy = Required(float, min=0.0, max=1.0)
    instrumentalness = Required(float, min=0.0, max=1.0)
    key = Required(int, min=0, max=11)
    liveness = Required(float, min=0.0, max=1.0)
    loudness = Required(float, min=-60.0, max=0.0)
    mode = Required(bool)
    speechiness = Required(float, min=0.0, max=1.0)
    tempo = Required(float, min=0, max=1000)
    time_signature = Required(int)
    valence = Required(float, min=0.0, max=1.0)

    def to_dict(self, *args, **kwargs):  # pylint: disable=arguments-differ
        _dict = super().to_dict(*args, **kwargs)
        if "mode" in _dict:
            _dict["mode"] = int(_dict["mode"])
        return _dict

    @classmethod
    def from_dict(cls, track):
        return cls(
            id=track["id"],
            acousticness=track["acousticness"],
            danceability=track["danceability"],
            duration_ms=track["duration_ms"],
            energy=track["energy"],
            instrumentalness=track["instrumentalness"],
            key=track["key"],
            liveness=track["liveness"],
            loudness=track["loudness"],
            mode=bool(track["mode"]),
            speechiness=track["speechiness"],
            tempo=track["tempo"],
            time_signature=track["time_signature"],
            valence=track["valence"],
        )


if config.database.connection.filename:
    config.database.connection.filename = os.path.expandvars(
        config.database.connection.filename
    )
db.bind(**config.database.connection)

GENERATE_MAPPING = os.getenv("SPFY_GENERATE_MAPPING")
if config.database.generate_mapping and (
    GENERATE_MAPPING is None or GENERATE_MAPPING == "true"
):
    db.generate_mapping(create_tables=True)



__version__ = "3.7.10"

import pathlib  # isort:skip

root = pathlib.Path(__file__).parent  # isort:skip
APP_NAME = "SPFY"

import os  # isort:skip

import kick  # isort:skip

kick.start(APP_NAME.lower())  # isort:skip

from kick import config, logger  # isort:skip

Unsplash = None
if config.unsplash.auth.client_id:
    from unsplash import Auth, Api

    auth = Auth(**config.unsplash.auth)
    Unsplash = Api(auth)
from .client import SpotifyClient
from .result import SpotifyResult
from .wrapper import Spotify
from .constants import *
from .exceptions import *



import numpy as _np
from cached_property import cached_property


class frange:
    """
    Return an object can be used to generate a generator or an array
    of floats from start (inclusive) to stop (exclusive) by step.
    This object stores the start, stop, step and length of
    the data. Uses less memory than storing a large array.
    Example
    -------
    An example of how to use this class to generate some data is
    as follows for some time data between 0 and 2 in steps of
    1e-3 (0.001)::
        you can create an frange object like so
        $ time = frange(0, 2, 1e-3)
        you can print the length of the array it will generate
        $ printlen(time) # prints length of frange, just like an array or list
        you can create a generator
        $ generator = time.generator # gets a generator instance
        $ for i in generator: # iterates through printing each element
        $     print(i)
        you can create a numpy array
        $ array = time.array # gets an array instance
        $ newarray = 5 * array # multiplies array by 5
        you can also get the start, stop and step by accessing the slice parameters
        $ start = time.start
        $ stop = time.stop
        $ step = time.step
    """

    def __init__(self, start, stop, step):
        """
        Intialises frange class instance. Sets start, top, step and
        len properties.
        Parameters
        ----------
        start : float
            starting point
        stop : float
            stopping point
        step : float
           stepping interval
        """
        self.slice = slice(start, stop, step)

    @property
    def start(self):
        return self.slice.start

    @property
    def step(self):
        return self.slice.step

    @property
    def stop(self):
        return self.slice.stop

    @cached_property
    def array(self):
        return self.get_array()

    @cached_property
    def generator(self):
        return self.get_generator()

    def __getitem__(self, idx):
        if idx == -1:
            return self.stop - self.step

        if idx == 0:
            return self.start

        return self.array[idx]

    def get_generator(self):
        """
        Returns a generator for the frange object instance.
        Returns
        -------
        gen : generator
            A generator that yields successive samples from start (inclusive)
            to stop (exclusive) in step steps.
        """
        return drange(self.start, self.stop, self.step)

    def get_array(self):
        """
        Returns an numpy array containing the values from start (inclusive)
        to stop (exclusive) in step steps.
        Returns
        -------
        array : ndarray
            Array of values from start (inclusive)
            to stop (exclusive) in step steps.
        """
        return _np.arange(self.start, self.stop, self.step)

    def __len__(self):
        return self.array.size

    def __iter__(self):
        return self.generator


def drange(start, stop, step):
    """
    A generator that yields successive samples from start (inclusive)
    to stop (exclusive) in step intervals.
    Parameters
    ----------
    start : float
        starting point
    stop : float
        stopping point
    step : float
        stepping interval
    Yields
    ------
    x : float
        next sample
    """
    x = start
    if step > 0:
        while x + step <= stop:  # produces same behaviour as numpy.arange
            yield x

            x += step
    elif step < 0:
        while x + step >= stop:  # produces same behaviour as numpy.arange
            yield x

            x += step
    else:
        raise ZeroDivisionError("Step must be non-zero")



import ujson as json


class SpotifyException(Exception):
    def __init__(
        self, *args, status_code=None, headers=None, text=None, url=None, **kwargs
    ):  # pylint: disable=unused-argument
        super().__init__()
        self.http_status_code = status_code
        self.headers = headers or {}
        if text and text != "null":
            try:
                response = json.loads(text)
                self.msg = f'{url}:\n {response["error"]["message"]}'
            except:
                self.msg = f"{url}: error"
        else:
            self.msg = f"{url}: error"

    def __str__(self):
        return f"""
        HTTP Status Code: {self.http_status_code}
        {self.msg}"""


class SpotifyRateLimitException(SpotifyException):
    def __init__(self, *args, retry_after=0, **kwargs):
        super().__init__(*args, **kwargs)
        self.retry_after = retry_after


class SpotifyForbiddenException(SpotifyException):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SpotifyDeviceUnavailableException(SpotifyException):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SpotifyCredentialsException(Exception):
    def __str__(self):
        return """
        You need to set your Spotify API credentials in ~/.config/spfy/config.toml

        Get your credentials at
            https://developer.spotify.com/my-applications
        """


class SpotifyAuthException(Exception):
    def __str__(self):
        return """
        You need to authenticate before making any request.
        """


class NoDatabaseConnection(Exception):
    pass



import pathlib

from setuptools import find_packages, setup

CONFIGDIR = pathlib.Path.home() / ".config" / "spfy"
CONFIGDIR.mkdir(parents=True, exist_ok=True)
with open("spfy/__init__.py", "r") as f:
    for line in f:
        if line.startswith("__version__"):
            version = line.strip().split("=")[1].strip(" '\"")
            break

    else:
        version = "0.0.1"
with open("README.md", "rb") as f:
    readme = f.read().decode("utf-8")
REQUIRES = [
    "addict",
    "aiohttp",
    "aioredis",
    "async_generator",
    "asyncpg",
    "backoff",
    "cachecontrol",
    "cached_property",
    "colorthief",
    "fire",
    "first",
    "gunicorn",
    "hug",
    "kick>=1.1.0",
    "lockfile",
    "mailer",
    "msgpack",
    "oauthlib",
    "pandas",
    "pony",
    "psycopg2-binary",
    "pycountry",
    "python-unsplash-async>=1.0.4",
    "urllib3<1.25,>=1.21.1",
    "requests",
    "requests_oauthlib",
    "tenacity",
    "ujson",
]
try:
    import sys

    if sys.platform == "Linux":
        REQUIRES.append("pyalsaaudio")
except:
    pass
setup(
    name="spfy",
    version=version,
    description="",
    long_description=readme,
    author="Alin Panaitiu",
    author_email="alin.p32@gmail.com",
    maintainer="Alin Panaitiu",
    maintainer_email="alin.p32@gmail.com",
    url="https://github.com/alin23/spfy",
    license="MIT/Apache-2.0",
    keywords=["spotify", "spfy"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    install_requires=REQUIRES,
    tests_require=["pytest"],
    packages=find_packages(),
    package_data={"spfy": ["config/*.toml", "html/*.html"]},
    data_files=[(str(CONFIGDIR), ["spfy/config/config.toml"])],
    entry_points={
        "console_scripts": [
            "spotify = spfy.wrapper:main",
            "spotify_async = spfy.asynch.wrapper:main",
        ]
    },
)



import re
from enum import Enum

from .frange import frange


class ItemType(Enum):
    TRACKS = "tracks"
    TRACK = "track"
    PLAYLIST = "playlist"
    ARTIST = "artist"
    ALBUM = "album"


class TimeRange(Enum):
    SHORT_TERM = "short_term"
    MEDIUM_TERM = "medium_term"
    LONG_TERM = "long_term"


class VolumeBackend(Enum):
    SPOTIFY = "spotify"
    APPLESCRIPT = "applescript"
    ALSA = "alsa"
    LINUX = "linux"


class AuthFlow(Enum):
    AUTHORIZATION_CODE = "authorization_code"
    CLIENT_CREDENTIALS = "client_credentials"


class AudioFeature(Enum):
    ACOUSTICNESS = "acousticness"
    DANCEABILITY = "danceability"
    DURATION_MS = "duration_ms"
    ENERGY = "energy"
    INSTRUMENTALNESS = "instrumentalness"
    KEY = "key"
    LIVENESS = "liveness"
    LOUDNESS = "loudness"
    MODE = "mode"
    POPULARITY = "popularity"
    SPEECHINESS = "speechiness"
    TEMPO = "tempo"
    TIME_SIGNATURE = "time_signature"
    VALENCE = "valence"


class AudioFeatureRange(Enum):
    ACOUSTICNESS = frange(0.0, 1.01, 0.01)
    DANCEABILITY = frange(0.0, 1.01, 0.01)
    DURATION_MS = range((30 * 60 * 1000) + 1)
    ENERGY = frange(0.0, 1.01, 0.01)
    INSTRUMENTALNESS = frange(0.0, 1.01, 0.01)
    KEY = range(0, 12)
    LIVENESS = frange(0.0, 1.01, 0.01)
    LOUDNESS = frange(-60.0, 0.1, 0.1)
    MODE = range(2)
    POPULARITY = range(101)
    SPEECHINESS = frange(0.0, 1.01, 0.01)
    TEMPO = range(321)
    TIME_SIGNATURE = range(9)
    VALENCE = frange(0.0, 1.01, 0.01)


class Scope(Enum):
    PLAYLIST_READ_PRIVATE = "playlist-read-private"
    PLAYLIST_READ_COLLABORATIVE = "playlist-read-collaborative"
    PLAYLIST_MODIFY_PUBLIC = "playlist-modify-public"
    PLAYLIST_MODIFY_PRIVATE = "playlist-modify-private"
    STREAMING = "streaming"
    UGC_IMAGE_UPLOAD = "ugc-image-upload"
    USER_FOLLOW_MODIFY = "user-follow-modify"
    USER_FOLLOW_READ = "user-follow-read"
    USER_LIBRARY_READ = "user-library-read"
    USER_LIBRARY_MODIFY = "user-library-modify"
    USER_READ_PRIVATE = "user-read-private"
    USER_READ_BIRTHDATE = "user-read-birthdate"
    USER_READ_EMAIL = "user-read-email"
    USER_TOP_READ = "user-top-read"
    USER_READ_PLAYBACK_STATE = "user-read-playback-state"
    USER_MODIFY_PLAYBACK_STATE = "user-modify-playback-state"
    USER_READ_CURRENTLY_PLAYING = "user-read-currently-playing"
    USER_READ_RECENTLY_PLAYED = "user-read-recently-played"


AllScopes = [scope.value for scope in Scope]


class API(Enum):
    PREFIX = "https://api.spotify.com"
    ALBUM = "/v1/albums/{id}"
    ALBUM_TRACKS = "/v1/albums/{id}/tracks"
    ALBUMS = "/v1/albums"
    ARTIST = "/v1/artists/{id}"
    ARTIST_ALBUMS = "/v1/artists/{id}/albums"
    ARTIST_RELATED_ARTISTS = "/v1/artists/{id}/related-artists"
    ARTIST_TOP_TRACKS = "/v1/artists/{id}/top-tracks"
    ARTISTS = "/v1/artists"
    AUDIO_ANALYSIS = "/v1/audio-analysis/{id}"
    AUDIO_FEATURES_SINGLE = "/v1/audio-features/{id}"
    AUDIO_FEATURES_MULTIPLE = "/v1/audio-features"
    RECOMMENDATIONS = "/v1/recommendations"
    RECOMMENDATIONS_GENRES = "/v1/recommendations/available-genre-seeds"
    AUTHORIZE = "https://accounts.spotify.com/authorize"
    TOKEN = "https://accounts.spotify.com/api/token"
    CATEGORIES = "/v1/browse/categories"
    CATEGORY = "/v1/browse/categories/{id}"
    CATEGORY_PLAYLISTS = "/v1/browse/categories/{id}/playlists"
    FEATURED_PLAYLISTS = "/v1/browse/featured-playlists"
    NEW_RELEASES = "/v1/browse/new-releases"
    ME = "/v1/me"
    MY_ALBUMS = "/v1/me/albums"
    MY_ALBUMS_CONTAINS = "/v1/me/albums/contains"
    MY_FOLLOWING = "/v1/me/following"
    MY_FOLLOWING_CONTAINS = "/v1/me/following/contains"
    MY_PLAYLISTS = "/v1/me/playlists"
    MY_TOP = "/v1/me/top/{type}"
    MY_TRACKS = "/v1/me/tracks"
    MY_TRACKS_CONTAINS = "/v1/me/tracks/contains"
    PLAYLIST = "/v1/users/{user_id}/playlists/{playlist_id}"
    PLAYLISTS = "/v1/users/{user_id}/playlists"
    PLAYLIST_FOLLOWERS = "/v1/users/{owner_id}/playlists/{playlist_id}/followers"
    PLAYLIST_FOLLOWERS_CONTAINS = (
        "/v1/users/{user_id}/playlists/{playlist_id}/followers/contains"
    )
    PLAYLIST_IMAGES = "/v1/users/{user_id}/playlists/{playlist_id}/images"
    PLAYLIST_TRACKS = "/v1/users/{user_id}/playlists/{playlist_id}/tracks"
    SEARCH_ALBUM = "/v1/search?type=album"
    SEARCH_ARTIST = "/v1/search?type=artist"
    SEARCH_PLAYLIST = "/v1/search?type=playlist"
    SEARCH_TRACK = "/v1/search?type=track"
    TRACK = "/v1/tracks/{id}"
    TRACKS = "/v1/tracks"
    USER = "/v1/users/{user_id}"
    CURRENTLY_PLAYING = "/v1/me/player/currently-playing"
    DEVICES = "/v1/me/player/devices"
    NEXT = "/v1/me/player/next"
    PAUSE = "/v1/me/player/pause"
    PLAY = "/v1/me/player/play"
    PLAYER = "/v1/me/player"
    PREVIOUS = "/v1/me/player/previous"
    RECENTLY_PLAYED = "/v1/me/player/recently-played"
    REPEAT = "/v1/me/player/repeat"
    SEEK = "/v1/me/player/seek"
    SHUFFLE = "/v1/me/player/shuffle"
    VOLUME = "/v1/me/player/volume"


VOLUME_FADE_SECONDS = 5 * 60
DEVICE_ID_RE = re.compile(r"[a-zA-Z0-9]{40}")
PLAYLIST_URI_RE = re.compile(r"spotify:user:[^:]+:playlist:[^:]+")
MANELISTI = {
    "2Ieszafc1unlRGyRmhGDFB",
    "2JoWWy2bVRC2bcx67BwILT",
    "2pepyBw5kRmhDaFkK0hiRQ",
    "2ZLpHjgeA6bEDlEVHhOOXL",
    "55xJmWjZC4Pa3FwEsMY28w",
    "3qf1tO2qoPVryMXg02WKRh",
    "3VnLLHKFN1ZB38FoJC9UEx",
    "6byfmUAqPBLK0owbuFEXlF",
    "3gnne9V6R0jmzs4ksnjNdy",
    "6Ujb9g9pljLl91qELsPZYo",
    "7aTSDkvSajML4o7CLlynwK",
    "0z0oRip2G1bkbX4spG9BpC",
    "2ekMCPMFstPliouB6rrGJK",
    "7unGBhZ0rI9iM76JLRFUCT",
    "1GXLkfzd7QjC8XIn4Dz1vp",
    "0gV4T1KXkmkrgTwQ7opgAN",
    "1I6I9IA5b9Q0H7JAmd19n8",
    "2L8Uf1Doaz8iJK1QKwh1I4",
    "3nEGbFFQiHXnjhbJVwzega",
    "1f2Y1i56FYEzsiRdrAYjsq",
    "06f5XN8rUFdabqdXn1Xuqg",
    "0TWTJophavu8mv6OF8st9q",
    "67zjnbtDq97ewfnbCq6oXw",
    "36Pv4nrbKXaEZHJinKrGDx",
    "2jjbkZe4cW2DJHTVZ3TVws",
    "5kqGOZA4rJho5gHdDTX5aZ",
    "0g5DbVhyQh959B7Z2ZYdPp",
    "1i6II3AaDfzer4QeNFuXne",
    "3Bu7CGd7oH8Ckitp0EsyZ5",
    "2SBpyLN4PD8DcPWlKeA4K8",
    "7gDHM1wBJAZOr69gU2o9oo",
    "4EocdGKc8KuWvNY8it8yBC",
    "4f9wGjBz9mNYQDuxP6KL56",
    "13zongmqs6nw7IB8qJUoyC",
    "7vEj3NqOilkYqgfbLCM2PT",
    "00kgsXttY3ZuxxXC6rUrZe",
    "10CpGdX2tyq1vGgf2JZaHL",
    "0xjHfnYTyL2yhSSdHpttsq",
    "18Z7SMNytEkKaiixfOX50w",
    "1Lbh8dcDjsa3BcE9LWyXyD",
    "0y7yHqqTeiK9xw4JRoTEHG",
    "2TrnHdJsXpzfxMFIZViwdI",
    "0nqyHPKioCjZ91eUcQ24SB",
    "3s5gg1dJCw5igSIo2DSVmQ",
    "3Mpgf08uPOZUYMnFDRSRV8",
    "5TZfuTwYa8JJga6TQ66vCG",
    "3LxWFtPzHcmzMO3k4rFUf7",
    "7noZM8YzX3rCq0uHcAuBXh",
    "7LLJvMSmGCDzaFKzGJVJ1V",
    "0fIp0Ar4hS7kBJjlv0dbEs",
    "7jSeLyH6h7qlWMaczER5Cd",
    "7iwW4N99bE6IqmetoSwJ7h",
    "1RNezj0YMlyMErtT86iWwg",
    "3VPPqsKpzzAFfY0f4qeaMy",
    "4Dk2VpQboT3GZ48989H9Sg",
    "3BPyexYwvn1oac4pCumPv8",
    "12bZ3FVNkc52pKG4YSgbSu",
    "7hyaaCD9Ydxt0dRJZaL4wt",
    "4EO93NpDQGwrdlzlQfnDCW",
    "6nU9Rg6157nS2p7zUfZJ6p",
    "2D0PmICj8cPJCN2KwdaFda",
    "0vXdmVC1505NYuTDouJo6G",
    "0fvOPoGXlSHNkGT2O0ATB7",
    "13Uivonlj8WicABhVkLyzl",
    "1WgCdu3YOI6rVp6Hr4NSc2",
    "3D18mEm1VBL6UMjklDeKvr",
    "5r9KPAzF16iwXfh0t15796",
    "0CisGgglR3QoT8r9LwaNNK",
    "3aBSSgGns3zM1UlXLuSWpL",
    "3nZ1cE2XuL2GaZ5VENV70S",
    "3FxwMafw6ORSgJgcm1BaaR",
    "2wtyDAPYMH93u49gBfVYPx",
    "7jct42t9M1sP3Mh83LX2mE",
    "4pp9qALPcnQmVpibuDyrQJ",
    "3ou1VcyERvmAyF1R7XIGmn",
    "4ouqMiepI4OQyd72R59pMQ",
    "7bQDLDdhEHr1ytwmKqTqgA",
    "0ksMyhTvlPBVBxpefKlxMC",
    "6ixTSq8mQsI26zrlUBTLS4",
    "6nvH1YG6WAFxGIEtX3JpzZ",
    "3axETwb6iK7HozXqv4bzqq",
    "2bqKyDcBvcWSbmKwhgF6Lq",
    "36tFKRhXaFBlycxiOvJVbs",
    "5KneFAP1oboQ4HpupQ2Y8m",
    "2KlIfHQACY0wbZt9TzKywN",
    "1lhvoeeajvhGmW4JEZ5nJv",
    "7BssMhtelPJEK1qycRG3iA",
    "2phUIFr6Q7b2ifSc4wJTOF",
    "1P9DU63Z7cYemiEjjeQmFs",
    "3Mpgf08uPOZUYMnFDRSRV8",
    "5tc4sgBPwnW8MfIceXu7VB",
    "39QykPFzm2emgtN9XyCJmV",
    "6I45uNgBW2URvWbQKqU5HY",
    "52OT0kDJk0eHbGJoXVXoID",
    "3MOeFK0Be57Mi1x5oLCfuL",
    "5VjP62EoivYMUpx5i6O55W",
    "4uNuwIueJrsSRtAXgRzDmb",
    "416ojj5IWmW6UJOrBDo931",
    "3Ols0qSo3Jy6JXsuzpFvIw",
    "0bEgGdcKWtP80aFtWTN5du",
    "3LMJ8PwGKLVgQiXNaTjzfD",
    "4sbgJi3VL7O1vS381MsLKn",
    "4RS1TcSqxr45eMFQe5wxdn",
    "1SGx0Rio4xIMWvgOLU4h3E",
    "4KBDKBcZsrqI7omC9s1KlG",
    "1ghEvzFzgDEpXsJTikumGw",
    "1zA7dHMrodgGqEN8owgAe3",
    "2EzZEVmCMtOSJtC8i1WeOQ",
    "09CTGylU0khFHdw7fMjyWg",
    "1pmqhKz2NUq8FEiX00brLu",
    "0j1HQ4RWuFbJPx2LnIITmH",
    "4tAxFVQ9LN2qVEi41ilqdE",
    "6XuwlE6QNkhGKauGl4jM7t",
    "034IOMKjllnYpbg9gqmavp",
    "14Hi4xJIidPV5v3RqEGLKV",
    "4ttfBkWrRYxGvHb9jcN9SF",
    "4KKWPAUUkcouh7xKjhLyLt",
    "3JI7vQAX2z1B0PxX01JJRp",
    "3Bq7FLSOaiIikbW6dawZNR",
    "4WwEX3orWGm7Fu0GyHTZBP",
    "6BqA8d5VxPy4nrk0TrKJNf",
    "0WYeg1DLQThOv5edYVaAeF",
    "1RbiY7VrOLuxuxTgGuUznZ",
    "79E5wmHfOMwxy4WCh4mDIL",
    "3BIuTXBzsorpOEIZVhwTss",
    "7sdwE7oJbhmsjjYeDYUQ3P",
    "2tqwGlXSGYYuZvBjF4wpuj",
    "0LRizmzjNKAjWlmkYBJi8V",
    "52fwrx4YWimHFpHqQCcLBb",
    "7KL6Xmrib9qHpU1xD9gRPB",
    "0EIBqLeqCKubCI2ZYYQDRw",
    "7c3NeCsxUjVOGVqaMOMmVP",
    "2KLd63rASWHMFH1jDp6Uoz",
    "4hlkJZ5XzNFKqxBgP0D7GB",
    "4ZknxfV6pMMTWCESLnra6X",
    "78Dw4QMPizHHRjcJSvgnYB",
    "4egN2mVCQGfz9NUuMBIjkg",
    "100vi1IPH5Go3Pb2UM3BWp",
    "6rESRpDyAmxiU41eboLcIR",
    "6JHsPCTCGVnvwEdsqdDgK6",
    "7xl58ylWDt598otFvFNFB0",
    "1HXo00S8hutdzcPZmhchLP",
    "5GxeTUY8mjqd8HczZqj2aI",
}
